from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor.opensslconfigcheck import (
    check_crypto_policies,
    check_default_modules,
    check_duplicate_extensions,
    check_min_max_protocol
)
from leapp.libraries.stdlib import api
from leapp.models import OpenSslConfig, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class OpenSslConfigCheck(Actor):
    """
    The OpenSSL configuration changed between RHEL8 and RHEL9 significantly with the rebase to
    OpenSSL 3.0. There are several things to check:

     * If the file was not modified by user, the RPM will take care of the upgrade
     * The file was modified and for some reason the link to the system-wide crypto
       policies is missing -- this is not recommended so we should warn users about that
     * The new OpenSSL 3.0 is using providers so we need to add them to the configuration
       file to make sure they behave the same way as when the original configuration file
       is used, especially for FIPS mode.
    """

    name = 'open_ssl_config_check'
    consumes = (OpenSslConfig,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag,)

    def process(self):
        openssl_messages = self.consume(OpenSslConfig)
        config = next(openssl_messages, None)
        if list(openssl_messages):
            api.current_logger().warning('Unexpectedly received more than one OpenSslConfig message.')
        if not config:
            raise StopActorExecutionError(
                'Could not check openssl configuration', details={'details': 'No OpenSslConfig facts found.'}
            )

        # If the configuration file was not modified, the rpm update will bring the new
        # changes by itself
        if not config.modified:
            return

        # Missing crypto policies is not recommended
        check_crypto_policies(config)

        # MinProtocol and MaxProtocol in [tls_system_default] changed their meaning
        check_min_max_protocol(config)

        # If the configuration file contains several X509 extensions with the same name,
        # only the last one will be used.
        check_duplicate_extensions(config)

        # Check and report what we are going to rewrite.
        #
        # Change the initialization:
        #
        # - openssl_conf = default_modules
        # + openssl_conf = openssl_init
        #
        # Rename the default block and link the providers block:
        #
        # - [default_modules]
        # + [openssl_init]
        # + providers = provider_sect
        #
        # Add the providers block:
        #
        # + [provider_sect]
        # + default = default_sect
        # + ##legacy = legacy_sect
        # +
        # + [default_sect]
        # + activate = 1
        # +
        # + ##[legacy_sect]
        # + ##activate = 1
        check_default_modules(config)



import os
import re

from leapp.libraries.actor import pluginscanner
from leapp.libraries.common.config.version import get_source_major_version
from leapp.libraries.stdlib import api
from leapp.models import PkgManagerInfo

YUM_CONFIG_PATH = '/etc/yum.conf'
DNF_CONFIG_PATH = '/etc/dnf/dnf.conf'


def _get_releasever_path():
    default_manager = 'yum' if get_source_major_version() == '7' else 'dnf'
    return '/etc/{}/vars/releasever'.format(default_manager)


def _releasever_exists(releasever_path):
    return os.path.isfile(releasever_path)


def get_etc_releasever():
    """ Get release version from "/etc/{yum,dnf}/vars/releasever" file """

    releasever_path = _get_releasever_path()
    if not _releasever_exists(releasever_path):
        return None

    with open(releasever_path, 'r') as fo:
        # we care about the first line only
        releasever = fo.readline().strip()

    return releasever


def _get_config_contents(config_path):
    if os.path.isfile(config_path):
        with open(config_path, 'r') as config:
            return config.read()
    return ''


def _get_proxy_if_set(manager_config_path):
    """
    Get proxy address from specified package manager config.

    :param str manager_config_path: path to a package manager config
    :returns: proxy address or None when not set
    :rtype: str
    """

    config = _get_config_contents(manager_config_path)

    for line in config.split('\n'):
        if re.match('^proxy[ \t]*=', line):
            proxy_address = line.split('=', 1)[1]
            return proxy_address.strip()

    return None


def get_configured_proxies():
    """
    Get a list of proxies used in dnf and yum configuration files.

    :returns: sorted list of unique proxies
    :rtype: List
    """

    configured_proxies = set()
    for config_path in (DNF_CONFIG_PATH, YUM_CONFIG_PATH):
        proxy = _get_proxy_if_set(config_path)
        if proxy:
            configured_proxies.add(proxy)

    return sorted(configured_proxies)


def process():
    pkg_manager_info = PkgManagerInfo()
    pkg_manager_info.etc_releasever = get_etc_releasever()
    pkg_manager_info.configured_proxies = get_configured_proxies()
    pkg_manager_info.enabled_plugins = pluginscanner.scan_enabled_package_manager_plugins()

    api.produce(pkg_manager_info)



import mock
import pytest
import six

from leapp.libraries.actor import nisscan
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import NISConfig

# Examples of /etc/yp.conf configuration file (default and configured)
YPBIND_DEFAULT_CONF = """# /etc/yp.conf - ypbind configuration file
# Valid entries are
#
# domain NISDOMAIN server HOSTNAME
#	Use server HOSTNAME for the domain NISDOMAIN"""

YPBIND_CONFIGURED_CONF = """# /etc/yp.conf - ypbind configuration file
domain whoisredhat.redhat server prod-db

domain whatisredhat.redhat server prod-db"""


@pytest.mark.parametrize('pkgs_installed', [
    ('ypserv',),
    ('ypbind',),
    ('ypserv-not-ypbind',),
    ('ypbind', 'ypserv'),
])
@pytest.mark.parametrize("fill_conf_file", [True, False])
@pytest.mark.parametrize("fill_ypserv_dir", [True, False])
def test_actor_nisscan(monkeypatch, pkgs_installed, fill_conf_file, fill_ypserv_dir):
    """
    Parametrized helper function for test_actor_* functions.

    Run the actor fed with our mocked functions and assert
    produced messages according to set arguments.

    Parameters:
        pkgs_installed  (tuple): installed pkgs
        fill_conf_file  (bool): not default ypbind config file
        fill_ypserv_dir  (bool): not default ypserv dir content
    """

    # Store final list of configured NIS packages
    configured_pkgs = []

    # Fill ypbind config
    yp_conf_content = YPBIND_CONFIGURED_CONF if fill_conf_file else YPBIND_DEFAULT_CONF

    # Fill ypserv dir files
    yp_dir_content = (nisscan.YPSERV_DEFAULT_FILES + ('example.com',) if fill_ypserv_dir
                      else nisscan.YPSERV_DEFAULT_FILES)

    # Mock 'isfile' & 'isdir' based on installed pkgs
    mocked_isfile = 'ypbind' in pkgs_installed
    mocked_isdir = 'ypserv' in pkgs_installed

    mock_config = mock.mock_open(read_data=yp_conf_content)
    with mock.patch("builtins.open" if six.PY3 else "__builtin__.open", mock_config):
        curr_actor_mocked = CurrentActorMocked()
        monkeypatch.setattr(api, 'current_actor', curr_actor_mocked)
        monkeypatch.setattr(api, "produce", produce_mocked())
        monkeypatch.setattr(nisscan.os, 'listdir', lambda dummy: yp_dir_content)
        monkeypatch.setattr(nisscan.os.path, 'isfile', lambda dummy: mocked_isfile)
        monkeypatch.setattr(nisscan.os.path, 'isdir', lambda dummy: mocked_isdir)

        # Executed actor fed with mocked functions
        nisscan.NISScanLibrary().process()

    # Filter NIS pkgs
    filtered_installed_pkgs = [x for x in pkgs_installed if x in nisscan.PACKAGES_NAMES]

    # Create correct list of pkgs for assert check
    for pkg in filtered_installed_pkgs:
        if (pkg == 'ypserv' and fill_ypserv_dir) or (pkg == 'ypbind' and fill_conf_file):
            configured_pkgs.append(pkg)

    # Sort NISConfig objects
    nisconf_template = set(NISConfig(nis_not_default_conf=configured_pkgs).nis_not_default_conf)
    nisconf_result = set(api.produce.model_instances[0].nis_not_default_conf)

    assert nisconf_template == nisconf_result



from leapp import reporting
from leapp.actors import Actor
from leapp.models import DistributionSignedRPM
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckPostfix(Actor):
    """
    Check if postfix is installed, check whether configuration update is needed.
    """

    name = 'check_postfix'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        for fact in self.consume(DistributionSignedRPM):
            for rpm in fact.items:
                if rpm.name == 'postfix':
                    create_report([
                        reporting.Title('Postfix has incompatible changes in the next major version'),
                        reporting.Summary(
                            'Postfix 3.x has so called "compatibility safety net" that runs Postfix programs '
                            'with backwards-compatible default settings. It will log a warning whenever '
                            'backwards-compatible default setting may be required for continuity of service. '
                            'Based on this logging the system administrator can decide if any '
                            'backwards-compatible settings need to be made permanent in main.cf or master.cf, '
                            'before turning off the backwards-compatibility safety net.\n'
                            'The backward compatibility safety net is by default turned off in Red Hat '
                            'Enterprise Linux 8.\n'
                            'It can be turned on by running:  "postconf -e compatibility_level=0\n'
                            'It can be turned off by running: "postconf -e compatibility_level=2\n\n'
                            'In the Postfix MySQL database client, the default "option_group" value has changed '
                            'to "client", i.e. it now reads options from the [client] group from the MySQL '
                            'configuration file. To disable it, set "option_group" to the empty string.\n\n'
                            'The postqueue command no longer forces all message arrival times to be reported '
                            'in UTC. To get the old behavior, set TZ=UTC in main.cf:import_environment.\n\n'
                            'Postfix 3.2 enables elliptic curve negotiation. This changes the default '
                            'smtpd_tls_eecdh_grade setting to "auto", and introduces a new parameter '
                            '"tls_eecdh_auto_curves" with the names of curves that may be negotiated.\n\n'
                            'The "master.cf" chroot default value has changed from "y" (yes) to "n" (no). '
                            'This applies to master.cf services where chroot field is not explicitly '
                            'specified.\n\n'
                            'The "append_dot_mydomain" default value has changed from "yes" to "no". You may '
                            'need changing it to "yes" if senders cannot use complete domain names in e-mail '
                            'addresses.\n\n'
                            'The "relay_domains" default value has changed from "$mydestination" to the empty '
                            'value. This could result in unexpected "Relay access denied" errors or ETRN errors, '
                            'because now will postfix by default relay only for the localhost.\n\n'
                            'The "mynetworks_style" default value has changed from "subnet" to "host". '
                            'This parameter is used to implement the "permit_mynetworks" feature. The change '
                            'could result in unexpected "access denied" errors, because postfix will now by '
                            'default trust only the local machine, not the remote SMTP clients on the '
                            'same IP subnetwork.\n\n'
                            'Postfix now supports dynamically loaded database plugins. Plugins are shipped '
                            'in individual RPM sub-packages. Correct database plugins have to be installed, '
                            'otherwise the specific database client will not work. For example for PostgreSQL '
                            'map to work, the postfix-pgsql RPM package has to be installed.\n',
                        ),
                        reporting.Severity(reporting.Severity.LOW),
                        reporting.Groups([reporting.Groups.SERVICES, reporting.Groups.EMAIL]),
                        reporting.RelatedResource('package', 'postfix')
                    ])
                    return



import json
import os

import pytest
import requests

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import repositoriesmapping
from leapp.libraries.common import fetch
from leapp.libraries.common.config import architecture, version
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import ConsumedDataAsset, PESIDRepositoryEntry, RPM

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def adjust_cwd():
    previous_cwd = os.getcwd()
    os.chdir(os.path.join(CUR_DIR, "./"))
    yield
    os.chdir(previous_cwd)


def test_scan_existing_valid_data(monkeypatch, adjust_cwd):
    """
    Tests whether an existing valid repomap file is loaded correctly.
    """

    with open('files/repomap_example.json') as repomap_file:
        data = json.load(repomap_file)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', dst_ver='8.4'))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    repositoriesmapping.scan_repositories(lambda dummy: data)

    assert api.produce.called, 'Actor did not produce any message when deserializing valid repomap data.'

    fail_description = 'Actor produced multiple messages, but only one was expected.'
    assert len(api.produce.model_instances) == 1, fail_description

    repo_mapping = api.produce.model_instances[0]

    # Verify that the loaded JSON data is matching the repomap file content
    # 1. Verify src_pesid -> target_pesids mappings are loaded and filtered correctly
    fail_description = 'Actor produced more mappings than there are source system relevant mappings in the test file.'
    assert len(repo_mapping.mapping) == 1, fail_description
    fail_description = 'Actor failed to load IPU-relevant mapping data correctly.'
    assert repo_mapping.mapping[0].source == 'pesid1', fail_description
    assert set(repo_mapping.mapping[0].target) == {'pesid2', 'pesid3'}, fail_description

    # 2. Verify that only repositories valid for the current IPU are produced
    pesid_repos = repo_mapping.repositories
    fail_description = 'Actor produced incorrect number of IPU-relevant pesid repos.'
    assert len(pesid_repos) == 3, fail_description

    expected_pesid_repos = [
        PESIDRepositoryEntry(
            pesid='pesid1',
            major_version='7',
            repoid='some-rhel-7-repoid',
            arch='x86_64',
            repo_type='rpm',
            channel='eus',
            rhui=''
        ),
        PESIDRepositoryEntry(
            pesid='pesid2',
            major_version='8',
            repoid='some-rhel-8-repoid1',
            arch='x86_64',
            repo_type='rpm',
            channel='eus',
            rhui=''
        ),
        PESIDRepositoryEntry(
            pesid='pesid3',
            major_version='8',
            repoid='some-rhel-8-repoid2',
            arch='x86_64',
            repo_type='rpm',
            channel='eus',
            rhui=''
        ),
    ]

    fail_description = 'Expected pesid repo is not present in the deserialization output.'
    for expected_pesid_repo in expected_pesid_repos:
        assert expected_pesid_repo in pesid_repos, fail_description


def test_scan_repositories_with_missing_data(monkeypatch):
    """
    Tests whether the scanning process fails gracefully when no data are read.
    """
    mocked_actor = CurrentActorMocked(src_ver='7.9', dst_ver='8.4', msgs=[])

    # Patch the mocked actor as the library will verify caller/callee contract
    mocked_actor.produces = (ConsumedDataAsset, )

    monkeypatch.setattr(api, 'current_actor', mocked_actor)
    monkeypatch.setattr(api, 'produce', produce_mocked())

    def read_or_fetch_mocked(*args, **kwargs):
        return ''

    monkeypatch.setattr(fetch, 'read_or_fetch', read_or_fetch_mocked)

    with pytest.raises(StopActorExecutionError) as missing_data_error:
        repositoriesmapping.scan_repositories()
    assert 'does not contain a valid JSON object' in str(missing_data_error)


def test_scan_repositories_with_empty_data(monkeypatch):
    """
    Tests whether the scanning process fails gracefully when empty json data received.
    """

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', dst_ver='8.4'))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    with pytest.raises(StopActorExecutionError) as empty_data_error:
        repositoriesmapping.scan_repositories(lambda dummy: {})
    assert 'the JSON is missing a required field' in str(empty_data_error)


@pytest.mark.parametrize('version_format', ('0.0.0', '1.0.1', '1.1.0', '2.0.0'))
def test_scan_repositories_with_bad_json_data_version(monkeypatch, version_format):
    """
    Tests whether the json data is checked for the version field and error is raised if the version
    does not match the latest one.
    """

    json_data = {
        'datetime': '202107141655Z',
        'version_format': version_format,
        'mapping': [],
        'repositories': []
    }

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', dst_ver='8.4'))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    with pytest.raises(StopActorExecutionError) as bad_version_error:
        repositoriesmapping.scan_repositories(lambda dummy: json_data)

    assert 'mapping file is invalid' in str(bad_version_error)


def test_scan_repositories_with_mapping_to_pesid_without_repos(monkeypatch):
    """
    Tests that the loading of repositories mapping recognizes when there is a mapping with target pesid that does
    not have any repositories and inhibits the upgrade.
    """
    json_data = {
        'datetime': '202107141655Z',
        'version_format': repositoriesmapping.RepoMapData.VERSION_FORMAT,
        'mapping': [
            {
                'source_major_version': '7',
                'target_major_version': '8',
                'entries':  [
                    {
                        'source': 'source_pesid',
                        'target': ['nonexisting_pesid']
                    }
                ]
            }
        ],
        'repositories': [
            {
                'pesid': 'source_pesid',
                'entries': [
                    {
                        'major_version': '7',
                        'repoid': 'some-rhel-7-repo',
                        'arch': 'x86_64',
                        'repo_type': 'rpm',
                        'channel': 'eus'
                    }
                ]
            }
        ]
    }

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', dst_ver='8.4'))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    with pytest.raises(StopActorExecutionError) as error_info:
        repositoriesmapping.scan_repositories(lambda dummy: json_data)

    assert 'pesid is not related to any repository' in error_info.value.message


def test_scan_repositories_with_repo_entry_missing_required_fields(monkeypatch):
    """
    Tests whether deserialization of pesid repo entries missing some of the required fields
    is handled internally and StopActorExecutionError is propagated to the user.
    """

    json_data = {
        'datetime': '202107141655Z',
        'version_format': repositoriesmapping.RepoMapData.VERSION_FORMAT,
        'mapping': [
            {
                'source_major_version': '7',
                'target_major_version': '8',
                'entries':  [
                    {
                        'source': 'source_pesid',
                        'target': ['target_pesid']
                    }
                ]
            }
        ],
        'repositories': [
            {
                'pesid': 'source_pesid',
                'entries': [
                    {
                        'major_version': '7',
                        'repoid': 'some-rhel-9-repo1',
                        'arch': 'x86_64',
                    }
                ]
            },
            {
                'pesid': 'target_pesid',
                'entries': [
                    {
                        'major_version': '7',
                        'repoid': 'some-rhel-9-repo1',
                        'arch': 'x86_64',
                    }
                ]
            }
        ]
    }

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', dst_ver='8.4'))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    with pytest.raises(StopActorExecutionError) as error_info:
        repositoriesmapping.scan_repositories(lambda dummy: json_data)

    assert 'the JSON is missing a required field' in error_info.value.message


def test_scan_repositories_with_repo_entry_mapping_target_not_a_list(monkeypatch):
    """
    Tests whether deserialization of a mapping entry that has its target field set to a string
    is handled internally and StopActorExecutionError is propagated to the user.
    """

    json_data = {
        'datetime': '202107141655Z',
        'version_format': repositoriesmapping.RepoMapData.VERSION_FORMAT,
        'mapping': [
            {
                'source_major_version': '7',
                'target_major_version': '8',
                'entries':  [
                    {
                        'source': 'source_pesid',
                        'target': 'target_pesid'
                    }
                ]
            }
        ],
        'repositories': [
            {
                'pesid': 'source_pesid',
                'entries': [
                    {
                        'major_version': '7',
                        'repoid': 'some-rhel-9-repo1',
                        'arch': 'x86_64',
                    }
                ]
            },
            {
                'pesid': 'target_pesid',
                'entries': [
                    {
                        'major_version': '7',
                        'repoid': 'some-rhel-9-repo1',
                        'arch': 'x86_64',
                    }
                ]
            }
        ]
    }

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', dst_ver='8.4'))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    with pytest.raises(StopActorExecutionError) as error_info:
        repositoriesmapping.scan_repositories(lambda dummy: json_data)

    assert 'repository mapping file is invalid' in error_info.value.message



# -*- coding: utf-8 -*-
import os
import shutil
import tempfile

import pytest

from leapp.libraries.actor.rootscanner import scan_dir


@pytest.mark.parametrize("filename,symlink,count_invalid",
                         [(u'a_utf_file'.encode('utf-8'), u"utf8_symlink".encode('utf-8'), 0),
                          (u'простофайл'.encode('koi8-r'), u"этонеутф8".encode('koi8-r'), 2),
                          (u'a_utf_file'.encode('utf-8'), u"этонеутф8".encode('koi8-r'), 1)])
def test_invalid_symlinks(filename, symlink, count_invalid):
    # Let's create a directory with both valid utf-8 and non-utf symlinks
    # NOTE(ivasilev) As this has to run for python2 as well can't use the nice tempfile.TemporaryDirectory way
    tmpdirname = tempfile.mkdtemp()
    # create the file in the temp directory
    path_to_file = os.path.join(tmpdirname.encode('utf-8'), filename)
    path_to_symlink = os.path.join(tmpdirname.encode('utf-8'), symlink)
    with open(path_to_file, 'w') as f:
        f.write('Some data here')
    # create a symlink
    os.symlink(path_to_file, path_to_symlink)
    # run scan_dir
    model = scan_dir(tmpdirname.encode('utf-8'))
    # verify the results
    assert len(model.items) == 2 - count_invalid
    assert len(model.invalid_items) == count_invalid
    # cleanup
    shutil.rmtree(tmpdirname)



import errno
import os

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import utils
from leapp.libraries.stdlib import api
from leapp.models import NetworkManagerConnection, NetworkManagerConnectionProperty, NetworkManagerConnectionSetting

libnm_available = False
err_details = None
try:
    import gi
    try:
        gi.require_version("NM", "1.0")
        from gi.repository import GLib, NM
        libnm_available = True
    except ValueError:
        err_details = 'NetworkManager-libnm package is not available'
except ImportError:
    err_details = 'python3-gobject-base package is not available'

NM_CONN_DIR = "/etc/NetworkManager/system-connections"


def process_file(filename):
    # We're running this through libnm in order to clear the secrets.
    # We don't know what keys are secret, but libnm does.
    keyfile = GLib.KeyFile()
    keyfile.load_from_file(filename, GLib.KeyFileFlags.NONE)
    con = NM.keyfile_read(keyfile, NM_CONN_DIR, NM.KeyfileHandlerFlags.NONE)
    con.clear_secrets()
    keyfile = NM.keyfile_write(con, NM.KeyfileHandlerFlags.NONE)
    cp = utils.parse_config(keyfile.to_data()[0])

    settings = []
    for setting_name in cp.sections():
        properties = []
        for name, value in cp.items(setting_name, raw=True):
            properties.append(NetworkManagerConnectionProperty(name=name, value=value))
        settings.append(
            NetworkManagerConnectionSetting(name=setting_name, properties=properties)
        )
    api.produce(NetworkManagerConnection(filename=filename, settings=settings))


def process_dir(directory):
    try:
        keyfiles = os.listdir(directory)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return
        raise

    for f in keyfiles:
        process_file(os.path.join(NM_CONN_DIR, f))


def process():
    if libnm_available:
        process_dir(NM_CONN_DIR)
    else:
        raise StopActorExecutionError(
            message='Failed to read NetworkManager connections',
            details=err_details
            )



import errno
import itertools
import os
import shutil
from collections import namedtuple

from leapp.libraries.common.config import get_all_envs
from leapp.libraries.common.config.version import get_source_major_version
from leapp.libraries.stdlib import api, CalledProcessError, run

# Using ALWAYS_BIND will crash the upgrade process if the file does not exist.
# Consider instead adding an entry to the ScanFilesToCopyIntoTargetSystem actor that
# conditionally (only if it exists) creates CopyFile message to the TargetUserspaceCreator.
ALWAYS_BIND = []

ErrorData = namedtuple('ErrorData', ['summary', 'details'])


class MountingMode(object):
    """
    MountingMode are types of mounts supported by the library
    """
    BIND = 'bind'
    """ Used for bind mounts """
    OVERLAY = 'overlay'
    """ Used for overlayfs mounts """
    LOOP = 'loop'
    """ Used for loop mounts """
    FSTYPE = 'fstype'
    """ Used to mount specific filesystem types such as procfs, sysfs etc """
    NONE = 'none'
    """ Used when no actual mount call needs to be issued """


def _makedirs(path, mode=0o777, exists_ok=True):
    """ Helper function which extends os.makedirs with exists_ok on all versions of python. """
    try:
        os.makedirs(path, mode=mode)
    except OSError:
        if not exists_ok or not os.path.isdir(path):
            raise


class MountError(Exception):
    """ Exception that is thrown when a mount related operation failed """

    def __init__(self, message, details):
        super(MountError, self).__init__(message)
        self.details = details


class IsolationType(object):
    """ Implementations for the different isolated actions types """
    class _Implementation(object):
        """ Base class for all isolated actions """

        def __init__(self, target, **kwargs):
            self.target = target

        def create(self):
            """ Create the isolation context """
            pass

        def close(self):
            """ Release the isolation context """
            pass

        def make_command(self, cmd):
            """ Transform the given command to the isolated environment """
            return cmd

    class NSPAWN(_Implementation):
        """ systemd-nspawn implementation """

        def __init__(self, target, binds=(), env_vars=None):
            super(IsolationType.NSPAWN, self).__init__(target=target)
            self.binds = list(binds) + ALWAYS_BIND
            self.env_vars = env_vars or get_all_envs()

        def make_command(self, cmd):
            """ Transform the command to be executed with systemd-nspawn """
            binds = ['--bind={}'.format(bind) for bind in self.binds]
            setenvs = ['--setenv={}={}'.format(env.name, env.value) for env in self.env_vars]
            final_cmd = ['systemd-nspawn', '--register=no', '--quiet']
            if get_source_major_version() != '7':
                # TODO: check whether we could use the --keep unit on el7 too.
                # in such a case, just add line into the previous solution..
                # TODO: the same about --capability=all
                final_cmd += ['--keep-unit', '--capability=all']
            return final_cmd + ['-D', self.target] + binds + setenvs + cmd

    class CHROOT(_Implementation):
        """ chroot implementation """

        def __init__(self, target):
            super(IsolationType.CHROOT, self).__init__(target)
            self.context = None

        def create(self):
            """ Create the necessary context for chroot based isolation """
            self.close()
            self.context = self._create_context()
            next(self.context)

        def _create_context(self):
            """ This will mount /proc, /sys and /dev for chroot executions """
            with TypedMount('proc', 'proc', os.path.join(self.target, 'proc')):
                with TypedMount('sysfs', 'sys', os.path.join(self.target, 'sys')):
                    with BindMount('/dev', os.path.join(self.target, 'dev')):
                        yield

        def close(self):
            """ Releasing the context and perform unmounting """
            if self.context:
                next(self.context)
            self.context = None

        def make_command(self, cmd):
            """ Transform the command to be executed in the chrooted environment """
            return [
                'chroot', self.target
            ] + cmd

    class NONE(_Implementation):
        """ Execute the given commands and perform the given operations on the real system and not isolated. """


class IsolatedActions(object):
    """ This class allows to perform actions in a manner as if the given base_dir would be the current root """

    _isolated = True

    def __init__(self, base_dir, implementation, **kwargs):
        self.base_dir = base_dir
        self.type = implementation(base_dir, **kwargs)

    def __enter__(self):
        self.type.create()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.type.close()

    def full_path(self, path):
        """
        Transform the path given to be prefixed with the base_dir, to get the real path on the system.

        The function is secured, so it is not possible to return path outside
        of the self.base_dir directory.

        Example: self.base_dir = '/var/lib/leapp/scratch/userspace'
                 path = '/etc/yum.repos.d/redhat.repo'
                 The result would be: /var/lib/leapp/scratch/userspace/etc/yum.repos.d/redhat.repo
        """
        return os.path.join(self.base_dir, os.path.abspath(path).lstrip('/'))

    def open(self, path, *args, **kwargs):
        """
        Open the path given as if it would be the real system.

        The only difference between this function and the python builtin open is the fact that the path uses
        self.full_path to translate the passed path argument. All other arguments are passed through
        """
        return open(self.full_path(path), *args, **kwargs)

    def call(self, cmd, *args, **kwargs):
        """ Running the given command using the leapp.libraries.stdlib.run function in a isolated manner. """
        return run(self.type.make_command(cmd), *args, **kwargs)

    def remove(self, path):
        """
        Removes the given file as it would be on the real system.
        """
        os.unlink(self.full_path(path))

    def remove_tree(self, path):
        """
        Removes the given directory recursively inside the isolated environment
        as it would be on the real system.

        If the destination doesn't exist, nothing happens.
        """
        try:
            shutil.rmtree(self.full_path(path))
        except EnvironmentError as e:
            # this is recommended way to handle it in Py2 & Py3
            if e.errno != errno.ENOENT:
                raise

    def mkdir(self, path, mode=0o777):
        """
        Creates the given path as it would be on the real system.
        """
        os.mkdir(self.full_path(path), mode=mode)

    def makedirs(self, path, mode=0o777, exists_ok=True):
        """
        Creates the whole path recursively for any missing part.
        """
        _makedirs(path=self.full_path(path), mode=mode, exists_ok=exists_ok)

    def copytree_to(self, src, dst):
        """
        Recursively copy an entire directory tree rooted at src. The destination directory,
        named by dst, must not already exist; it will be created as well as missing parent directories.

        The destination directory is considered to be in the isolated environment.
        The source directory is considered to be on the current system root.
        """
        shutil.copytree(src, self.full_path(dst))

    def copytree_from(self, src, dst):
        """
        Recursively copy an entire directory tree rooted at src. The destination directory,
        named by dst, must not already exist; it will be created as well as missing parent directories.

        The destination directory is considered to be on the current system root.
        The source directory is considered to be in the isolated environment.
        """
        shutil.copytree(self.full_path(src), dst)

    def copy_to(self, src, dst):
        """
        Copy the file src to the file or directory dst. If dst is a directory, a file with the same basename
        as src is created (or overwritten) in the directory specified. Permission bits are copied. src and dst
        are path names given as strings.
        copy_to also attempts to preserve file metadata.

        The source is expected to be on the current system.
        The destination is expected to be in the isolated environment.
        """
        shutil.copy2(src, self.full_path(dst))

    def copy_from(self, src, dst):
        """
        Copy the file src to the file or directory dst. If dst is a directory, a file with the same basename
        as src is created (or overwritten) in the directory specified. Permission bits are copied. src and dst
        are path names given as strings.
        copy_to also attempts to preserve file metadata.

        The source is expected to be in the isolated environment.
        The destination is expected to be on the current system.
        """
        shutil.copy2(self.full_path(src), dst)

    @classmethod
    def is_isolated(cls):
        """
        Tell whether the context is isolated or not.

        All classes except NotIsolatedActions return True.
        """
        return cls._isolated


class ChrootActions(IsolatedActions):
    """ Isolation with chroot """

    def __init__(self, base_dir):
        super(ChrootActions, self).__init__(base_dir=base_dir, implementation=IsolationType.CHROOT)


class NspawnActions(IsolatedActions):
    """ Isolation with systemd-nspawn """

    def __init__(self, base_dir, binds=(), env_vars=None):
        super(NspawnActions, self).__init__(
            base_dir=base_dir, implementation=IsolationType.NSPAWN, binds=binds, env_vars=env_vars)


class NotIsolatedActions(IsolatedActions):
    """ Non isolated executed. """
    _isolated = False

    def __init__(self, base_dir):
        super(NotIsolatedActions, self).__init__(base_dir=base_dir, implementation=IsolationType.NONE)


class MountConfig(object):
    """ Options for Mount """
    _Options = namedtuple('_Options', ('should_create', 'should_cleanup'))
    AttachOnly = _Options(should_create=False, should_cleanup=False)
    """ Do not perform any mount operations, and do not clean up afterwards """
    Attach = _Options(should_create=False, should_cleanup=True)
    """ Do not perform any mount operations, however cleanup afterwards """
    MountOnly = _Options(should_create=True, should_cleanup=False)
    """ Create all necessary directories and perform mount calls, but do not cleanup afterwards """
    Mount = _Options(should_create=True, should_cleanup=True)
    """ Create all necessary directories and perform mount calls and cleanup afterwards """


class MountingBase(object):
    """ Base class for all mount operations """

    def __init__(self, source, target, mode, config=MountConfig.Mount):
        self._mode = mode
        self.source = source
        self.target = target
        self._config = config
        self.additional_directories = ()

    def _mount_options(self):
        """
        Options to use with the mount call, individual implementations may override this function to return the
        correct parameters
        """
        return ['-o', self._mode, self.source]

    def chroot(self):
        """ Create a ChrootActions instance for this mount """
        return ChrootActions(self.target)

    def nspawn(self):
        """ Create a NspawnActions instance for this mount """
        return NspawnActions(self.target)

    def real(self):
        """ Create a NotIsolatedActions instance for this mount """
        return NotIsolatedActions(self.target)

    def _cleanup(self):
        """ Cleanup operations """
        if os.path.exists(self.target) and os.path.ismount(self.target):
            try:
                run(['umount', '-fl', self.target], split=False)
            except (OSError, CalledProcessError) as e:
                api.current_logger().warning('Unmounting %s failed with: %s', self.target, str(e))
        for directory in itertools.chain(self.additional_directories, (self.target,)):
            try:
                run(['rm', '-rf', directory], split=False)
            except (OSError, CalledProcessError) as e:
                api.current_logger().warning('Removing mount directory %s failed with: %s', directory, str(e))

    def mount(self):
        """ Performs the mount if MountConfig.should_create = True """
        if self._config.should_create:
            self._create()

    def _create(self):
        self._cleanup()
        for directory in itertools.chain(self.additional_directories, (self.target,)):
            try:
                _makedirs(directory, exists_ok=True)
            except (OSError) as e:
                raise MountError('Failed to create mount target directory {}'.format(directory), str(e))
        try:
            run(['mount'] + self._mount_options() + [self.target], split=False)
        except (OSError, CalledProcessError) as e:
            api.current_logger().warning('Mounting %s failed with: %s', self.target, str(e), exc_info=True)
            raise MountError(
                message='Mount operation with mode {} from {} to {} failed: {}'.format(
                    self._mode, self.source, self.target, str(e)),
                details=None)
        return self

    def umount(self):
        """ Performs the umount if MountConfig.should_cleanup = True """
        if self._config.should_cleanup:
            self._cleanup()

    def __enter__(self):
        self.mount()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.umount()


class NullMount(MountingBase):
    """ This is basically a NoOp for compatibility with other mount operations, in case a mount is optional """

    def __init__(self, target, config=MountConfig.AttachOnly):
        super(NullMount, self).__init__(source=target, target=target, mode=MountingMode.NONE, config=config)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass


class LoopMount(MountingBase):
    """ Performs loop mounts """

    def __init__(self, source, target, config=MountConfig.Mount):
        super(LoopMount, self).__init__(source=source, target=target, mode=MountingMode.LOOP, config=config)


class BindMount(MountingBase):
    """ Performs bind mounts """

    def __init__(self, source, target, config=MountConfig.Mount):
        super(BindMount, self).__init__(source=source, target=target, mode=MountingMode.BIND, config=config)


class TypedMount(MountingBase):
    """ Performs a typed mounts """

    def __init__(self, fstype, source, target, config=MountConfig.Mount):
        super(TypedMount, self).__init__(source=source, target=target, mode=MountingMode.FSTYPE, config=config)
        self.fstype = fstype

    def _mount_options(self):
        return [
            '-t', self.fstype,
            self.source
        ]


class OverlayMount(MountingBase):
    """ Performs an overlayfs mount """

    def __init__(self, name, source, workdir, config=MountConfig.Mount):
        super(OverlayMount, self).__init__(source=source, target=os.path.join(workdir, name),
                                           mode=MountingMode.OVERLAY, config=config)
        self._upper_dir = os.path.join(workdir, 'upper')
        self._work_dir = os.path.join(workdir, 'work')
        self.additional_directories = (self._upper_dir, self._work_dir)

    def _mount_options(self):
        return [
            '-t', 'overlay', 'overlay2',
            '-o', 'lowerdir={},upperdir={},workdir={}'.format(self.source, self._upper_dir, self._work_dir)
        ]


def mount_upgrade_iso_to_root_dir(root_dir, target_iso):
    """
    Context manager mounting the target RHEL ISO into the system root residing at `root_dir`.

    If the `target_iso` is None no action is performed.

    :param root_dir: Path to a directory containing a system root.
    :type root_dir: str
    :param target_iso: Description of the ISO to be mounted.
    :type target_iso: Optional[TargetOSInstallationImage]
    :rtype: Optional[LoopMount]
    """
    if not target_iso:
        return NullMount(root_dir)

    mountpoint = target_iso.mountpoint[1:]  # Strip the leading / from the absolute mountpoint
    mountpoint_in_root_dir = os.path.join(root_dir, mountpoint)
    return LoopMount(source=target_iso.path, target=mountpoint_in_root_dir)



from leapp.actors import Actor
from leapp.libraries.actor import opensshprotocolcheck
from leapp.models import OpenSshConfig, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class OpenSshProtocolCheck(Actor):
    """
    Protocol configuration option was removed.

    Check the value of Protocol in OpenSSH server config file
    and warn about its deprecation if it is set. This option was removed
    in RHEL 7.4, but it might still be hanging around.
    """

    name = 'open_ssh_protocol'
    consumes = (OpenSshConfig, )
    produces = (Report, )
    tags = (ChecksPhaseTag, IPUWorkflowTag, )

    def process(self):
        opensshprotocolcheck.process(self.consume(OpenSshConfig))



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class XFSPresence(Model):
    """
    A generic message reporting mountpoints with XFS using ftype = 0
    """
    topic = SystemInfoTopic

    present = fields.Boolean(default=False)
    """ XFS is used on at least one mountpoint """

    without_ftype = fields.Boolean(default=False)
    """ At least one mountpoint is using XFS with ftype=0 """

    mountpoints_without_ftype = fields.List(fields.String(), default=[])
    """ List of mountpoints that have ftype=0 """



import os

from leapp.models import DistributionSignedRPM, IpaInfo, RPM
from leapp.snactor.fixture import current_actor_context

DEFAULT_CONF = "/etc/ipa/default.conf"
CLIENT_STATE = "/var/lib/ipa-client/sysrestore/sysrestore.state"
SERVER_STATE = "/var/lib/ipa/sysrestore/sysrestore.state"


def mock_rpm(name):
    return RPM(
        name=name,
        epoch="0",
        packager="Red Hat Inc.",
        version="4.6.0",
        release="1.el7",
        arch="x86_64",
        pgpsig="dummy",
    )


def mock_rpms(*names):
    return DistributionSignedRPM(items=[mock_rpm(name) for name in names])


def mock_os_path_isfile(overrides):
    def mocked_os_path_isfile(name):
        if name in overrides:
            return overrides[name]
        raise ValueError

    return mocked_os_path_isfile


def assert_ipa_info(infos, client, server):
    assert len(infos) == 1

    info = infos[0]
    assert info.has_client_package == client
    assert info.is_client_configured == client
    assert info.has_server_package == server
    assert info.is_server_configured == server


def test_client_configured(monkeypatch, current_actor_context):
    monkeypatch.setattr(
        "os.path.isfile",
        mock_os_path_isfile(
            {DEFAULT_CONF: True, CLIENT_STATE: True, SERVER_STATE: False}
        ),
    )

    rpms = mock_rpms("ipa-client")
    current_actor_context.feed(rpms)

    current_actor_context.run()

    infos = current_actor_context.consume(IpaInfo)
    assert_ipa_info(infos, True, False)


def test_server_configured(monkeypatch, current_actor_context):
    monkeypatch.setattr(
        "os.path.isfile",
        mock_os_path_isfile(
            {DEFAULT_CONF: True, CLIENT_STATE: True, SERVER_STATE: True}
        ),
    )

    rpms = mock_rpms("ipa-client", "ipa-server")
    current_actor_context.feed(rpms)

    current_actor_context.run()

    infos = current_actor_context.consume(IpaInfo)
    assert_ipa_info(infos, True, True)


def test_not_configured(monkeypatch, current_actor_context):
    monkeypatch.setattr(
        "os.path.isfile",
        mock_os_path_isfile(
            {DEFAULT_CONF: True, CLIENT_STATE: False, SERVER_STATE: False}
        ),
    )

    rpms = mock_rpms()
    current_actor_context.feed(rpms)
    current_actor_context.run()

    infos = current_actor_context.consume(IpaInfo)
    assert_ipa_info(infos, False, False)



from collections import namedtuple
from enum import Enum

import pytest

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import checkrhui as checkrhui_lib
from leapp.libraries.common import rhsm, rhui
from leapp.libraries.common.config import mock_configs, version
from leapp.libraries.common.rhui import mk_rhui_setup, RHUIFamily
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import (
    CopyFile,
    InstalledRPM,
    RequiredTargetUserspacePackages,
    RHUIInfo,
    RPM,
    RpmTransactionTasks,
    TargetRHUIPostInstallTasks,
    TargetRHUIPreInstallTasks,
    TargetRHUISetupInfo,
    TargetUserSpacePreupgradeTasks
)
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context

RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'


def mk_pkg(name):
    return RPM(name=name, version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
               pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')


def mk_setup_info():
    pre_tasks = TargetRHUIPreInstallTasks()
    post_tasks = TargetRHUIPostInstallTasks()
    return TargetRHUISetupInfo(preinstall_tasks=pre_tasks, postinstall_tasks=post_tasks)


def iter_known_rhui_setups():
    for upgrade_path, providers in rhui.RHUI_CLOUD_MAP.items():
        for provider_variant, variant_description in providers.items():
            src_clients = variant_description['src_pkg']
            if isinstance(src_clients, str):
                src_clients = {src_clients, }

            yield provider_variant, upgrade_path, src_clients


def mk_cloud_map(variants):
    upg_path = {}
    for variant_desc in variants:
        provider, desc = next(iter(variant_desc.items()))
        upg_path[provider] = desc
    return upg_path


@pytest.mark.parametrize(
    ('extra_pkgs', 'rhui_setups', 'expected_result'),
    [
        (
            ['client'],
            {RHUIFamily('provider'): [mk_rhui_setup(clients={'client'})]},
            RHUIFamily('provider')
        ),
        (
            ['client'],
            {RHUIFamily('provider'): [mk_rhui_setup(clients={'missing_client'})]},
            None
        ),
        (
            ['clientA', 'clientB'],
            {RHUIFamily('provider'): [mk_rhui_setup(clients={'clientB'})]},
            RHUIFamily('provider')
        ),
        (
            ['clientA', 'clientB'],
            {
                RHUIFamily('provider'): [mk_rhui_setup(clients={'clientA'})],
                RHUIFamily('provider+'): [mk_rhui_setup(clients={'clientA', 'clientB'})],
            },
            RHUIFamily('provider+')
        ),
        (
            ['client'],
            {
                RHUIFamily('providerA'): [mk_rhui_setup(clients={'client'})],
                RHUIFamily('providerB'): [mk_rhui_setup(clients={'client'})],
            },
            StopActorExecutionError
        ),
    ]
)
def test_determine_rhui_src_variant(monkeypatch, extra_pkgs, rhui_setups, expected_result):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9'))
    installed_pkgs = {'zip', 'zsh', 'bash', 'grubby'}.union(set(extra_pkgs))

    if expected_result and not isinstance(expected_result, RHUIFamily):  # An exception
        with pytest.raises(expected_result) as err:
            checkrhui_lib.find_rhui_setup_matching_src_system(installed_pkgs, rhui_setups)
        assert 'ambiguous' in str(err)
        return

    variant_setup_pair = checkrhui_lib.find_rhui_setup_matching_src_system(installed_pkgs, rhui_setups)
    if not expected_result:
        assert variant_setup_pair == expected_result
    else:
        variant = variant_setup_pair[0]
        assert variant == expected_result


@pytest.mark.parametrize(
    ('extra_pkgs', 'target_rhui_setup', 'should_inhibit'),
    [
        (['pkg'], mk_rhui_setup(leapp_pkg='pkg'), False),
        ([], mk_rhui_setup(leapp_pkg='pkg'), True),
    ]
)
def test_inhibit_on_missing_leapp_rhui_pkg(monkeypatch, extra_pkgs, target_rhui_setup, should_inhibit):
    installed_pkgs = set(['bash', 'zsh', 'zip'] + extra_pkgs)
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    checkrhui_lib.inhibit_if_leapp_pkg_to_access_target_missing(installed_pkgs,
                                                                RHUIFamily('rhui-variant'),
                                                                target_rhui_setup)
    assert bool(reporting.create_report.called) == should_inhibit


def are_setup_infos_eq(actual, expected):
    eq = True
    eq &= actual.enable_only_repoids_in_copied_files == expected.enable_only_repoids_in_copied_files
    eq &= actual.files_supporting_client_operation == expected.files_supporting_client_operation
    eq &= actual.preinstall_tasks.files_to_remove == expected.preinstall_tasks.files_to_remove
    eq &= actual.preinstall_tasks.files_to_copy_into_overlay == expected.preinstall_tasks.files_to_copy_into_overlay
    eq &= actual.postinstall_tasks.files_to_copy == expected.postinstall_tasks.files_to_copy
    return eq


@pytest.mark.parametrize(
    ('provider', 'should_mutate'),
    [
        (RHUIFamily(rhui.RHUIProvider.GOOGLE), True),
        (RHUIFamily(rhui.RHUIProvider.GOOGLE, variant=rhui.RHUIVariant.SAP), True),
        (RHUIFamily('azure'), False),
    ]
)
def test_google_specific_customization(provider, should_mutate):
    setup_info = mk_setup_info()
    checkrhui_lib.customize_rhui_setup_for_gcp(provider, setup_info)

    if should_mutate:
        assert setup_info != mk_setup_info()
    else:
        assert setup_info == mk_setup_info()


@pytest.mark.parametrize(
    ('rhui_family', 'target_major', 'should_mutate'),
    [
        (RHUIFamily(rhui.RHUIProvider.AWS), '8', False),
        (RHUIFamily(rhui.RHUIProvider.AWS), '9', True),
        (RHUIFamily(rhui.RHUIProvider.AWS, variant=rhui.RHUIVariant.SAP), '9', True),
        (RHUIFamily('azure'), '9', False),
    ]
)
def test_aws_specific_customization(monkeypatch, rhui_family, target_major, should_mutate):
    dst_ver = '{major}.0'.format(major=target_major)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(dst_ver=dst_ver))

    setup_info = mk_setup_info()
    checkrhui_lib.customize_rhui_setup_for_aws(rhui_family, setup_info)

    was_mutated = not are_setup_infos_eq(setup_info, mk_setup_info())
    assert should_mutate == was_mutated


def produce_rhui_info_to_setup_target(monkeypatch):
    source_rhui_setup = mk_rhui_setup(
        clients={'src_pkg'},
        leapp_pkg='leapp_pkg',
        mandatory_files=[('src_file1', '/etc'), ('src_file2', '/var')],
    )

    target_rhui_setup = mk_rhui_setup(
        clients={'target_pkg'},
        leapp_pkg='leapp_pkg',
        mandatory_files=[('target_file1', '/etc'), ('target_file2', '/var')],
    )

    monkeypatch.setattr(api, 'get_common_folder_path', lambda dummy: 'common_folder')
    monkeypatch.setattr(api, 'produce', produce_mocked())

    checkrhui_lib.produce_rhui_info_to_setup_target('provider', source_rhui_setup, target_rhui_setup)

    assert len(api.produce.model_instances) == 1

    rhui_info = api.produce.model_instances[0]
    assert rhui_info.provider == 'provider'
    assert rhui_info.src_client_pkg_names == ['src_pkg']
    assert rhui_info.target_client_pkg_names == ['target_pkg']

    setup_info = rhui_info.target_client_setup_info

    expected_copies = {
        ('common_folder/provider/target_file1', '/etc'),
        ('common_folder/provider/target_file2', '/var')
    }
    actual_copies = {(instr.src, instr.dst) for instr in setup_info.preinstall_tasks.files_to_copy_in}

    assert expected_copies == actual_copies

    assert not setup_info.postinstall_tasks.files_to_copy


def test_produce_rpms_to_install_into_target(monkeypatch):
    source_rhui_setup = mk_rhui_setup(clients={'src_pkg'}, leapp_pkg='leapp_pkg')
    target_rhui_setup = mk_rhui_setup(clients={'target_pkg'}, leapp_pkg='leapp_pkg')

    monkeypatch.setattr(api, 'produce', produce_mocked())

    checkrhui_lib.produce_rpms_to_install_into_target(source_rhui_setup, target_rhui_setup)

    assert len(api.produce.model_instances) == 2
    userspace_tasks, target_rpm_tasks = api.produce.model_instances[0], api.produce.model_instances[1]

    if isinstance(target_rpm_tasks, TargetUserSpacePreupgradeTasks):
        userspace_tasks, target_rpm_tasks = target_rpm_tasks, userspace_tasks

    assert 'target_pkg' in target_rpm_tasks.to_install
    assert 'src_pkg' in target_rpm_tasks.to_remove
    assert 'target_pkg' in userspace_tasks.install_rpms


@pytest.mark.parametrize('skip_rhsm', (True, False))
def test_inform_about_upgrade_with_rhui_without_no_rhsm(monkeypatch, skip_rhsm):
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: skip_rhsm)
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    checkrhui_lib.inform_about_upgrade_with_rhui_without_no_rhsm()

    assert bool(reporting.create_report.called) is not skip_rhsm


class ExpectedAction(Enum):
    NOTHING = 1  # Actor should not produce anything
    INHIBIT = 2
    PRODUCE = 3  # Actor should produce RHUI related info


# Scenarios to cover:
# 1. source client + NO_RHSM -> RPMs are produced, and setup info is produced
# 2. source client -> inhibit
# 3. leapp pkg missing -> inhibit
@pytest.mark.parametrize(
    ('extra_installed_pkgs', 'skip_rhsm', 'expected_action'),
    [
        (['src_pkg', 'leapp_pkg'], True, ExpectedAction.PRODUCE),  # Everything OK
        (['src_pkg', 'leapp_pkg'], False, ExpectedAction.INHIBIT),  # No --no-rhsm
        (['src_pkg'], True, ExpectedAction.INHIBIT),  # Missing leapp-rhui package
        ([], True, ExpectedAction.NOTHING)  # Not a RHUI system
    ]
)
def test_process(monkeypatch, extra_installed_pkgs, skip_rhsm, expected_action):
    known_setups = {
        RHUIFamily('rhui-variant'): [
            mk_rhui_setup(clients={'src_pkg'}, os_version='7'),
            mk_rhui_setup(clients={'target_pkg'}, os_version='8', leapp_pkg='leapp_pkg',
                          mandatory_files=[('file1', '/etc'), ('file2', '/var')]),
        ]
    }

    installed_pkgs = {'zip', 'kernel-core', 'python'}.union(set(extra_installed_pkgs))
    installed_pkgs = [mk_pkg(pkg_name) for pkg_name in installed_pkgs]
    installed_rpms = InstalledRPM(items=installed_pkgs)

    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', msgs=[installed_rpms]))
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: skip_rhsm)
    monkeypatch.setattr(rhui, 'RHUI_SETUPS', known_setups)

    checkrhui_lib.process()

    if expected_action == ExpectedAction.NOTHING:
        assert not api.produce.called
        assert not reporting.create_report.called
    elif expected_action == ExpectedAction.INHIBIT:
        assert not api.produce.called
        assert len(reporting.create_report.reports) == 1
    else:  # expected_action = ExpectedAction.PRODUCE
        assert not reporting.create_report.called
        assert len(api.produce.model_instances) == 3
        assert any(isinstance(pkg, RpmTransactionTasks) for pkg in api.produce.model_instances)
        assert any(isinstance(pkg, RHUIInfo) for pkg in api.produce.model_instances)
        assert any(isinstance(pkg, TargetUserSpacePreupgradeTasks) for pkg in api.produce.model_instances)


@pytest.mark.parametrize('is_target_setup_known', (False, True))
def test_unknown_target_rhui_setup(monkeypatch, is_target_setup_known):
    rhui_family = RHUIFamily('rhui-variant')
    known_setups = {
        rhui_family: [
            mk_rhui_setup(clients={'src_pkg'}, os_version='7'),
        ]
    }

    if is_target_setup_known:
        target_setup = mk_rhui_setup(clients={'target_pkg'}, os_version='8', leapp_pkg='leapp_pkg')
        known_setups[rhui_family].append(target_setup)

    installed_pkgs = {'zip', 'kernel-core', 'python', 'src_pkg', 'leapp_pkg'}
    installed_pkgs = [mk_pkg(pkg_name) for pkg_name in installed_pkgs]
    installed_rpms = InstalledRPM(items=installed_pkgs)

    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', msgs=[installed_rpms]))
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: True)
    monkeypatch.setattr(rhui, 'RHUI_SETUPS', known_setups)

    if is_target_setup_known:
        checkrhui_lib.process()
        assert api.produce.called
    else:
        with pytest.raises(StopActorExecutionError):
            checkrhui_lib.process()



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import targetinitramfsgenerator
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.utils.deprecation import suppress_deprecation

from leapp.models import (    # isort:skip
    InitrdIncludes,    # deprecated
    DracutModule, KernelModule, InstalledTargetKernelInfo, TargetInitramfsTasks)

FILES = ['/file1', '/file2', '/dir/subdir/subsubdir/file3', '/file4', '/file5']
MODULES = [
    ('moduleA', None),
    ('moduleB', None),
    ('moduleC', '/some/path/moduleC'),
    ('moduleD', '/some/path/moduleD'),
]
KERNEL_VERSION = '4.18.0'
NO_INCLUDE_MSG = 'No additional files or modules required to add into the target initramfs.'


def raise_call_error(args=None):
    raise CalledProcessError(message='A Leapp Command Error occurred.',
                             command=args,
                             result={
                                 'signal': None,
                                 'exit_code': 1,
                                 'pid': 0,
                                 'stdout': 'fake',
                                 'stderr': 'fake'
                             })


class RunMocked(object):

    def __init__(self, raise_err=False):
        self.called = 0
        self.args = []
        self.raise_err = raise_err

    def __call__(self, args):
        self.called += 1
        self.args = args
        if self.raise_err:
            raise_call_error(args)


def _ensure_list(data):
    return data if isinstance(data, list) else [data]


def gen_TIT(dracut_modules, kernel_modules, files):
    files = _ensure_list(files)

    dracut_modules = [DracutModule(name=i[0], module_path=i[1]) for i in _ensure_list(dracut_modules)]
    kernel_modules = [KernelModule(name=i[0], module_path=i[1]) for i in _ensure_list(kernel_modules)]

    return TargetInitramfsTasks(
        include_files=files,
        include_dracut_modules=dracut_modules,
        include_kernel_modules=kernel_modules,
    )


@suppress_deprecation(InitrdIncludes)
def gen_InitrdIncludes(files):
    return InitrdIncludes(files=_ensure_list(files))


def test_no_includes(monkeypatch):
    run_mocked = RunMocked()
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[]))
    monkeypatch.setattr(api, 'current_logger', logger_mocked())
    monkeypatch.setattr(targetinitramfsgenerator, 'run', run_mocked)

    targetinitramfsgenerator.process()
    assert NO_INCLUDE_MSG in api.current_logger.dbgmsg
    assert not run_mocked.called


TEST_CASES = [
    [
        gen_InitrdIncludes(FILES[0:3]),
        gen_InitrdIncludes(FILES[3:]),
    ],
    [
        gen_TIT([], [], FILES[0:3]),
        gen_TIT([], [], FILES[3:]),
    ],
    [
        gen_InitrdIncludes(FILES[0:3]),
        gen_TIT([], [], FILES[3:]),
    ],
]


@pytest.mark.parametrize('msgs', TEST_CASES)
def test_no_kernel_version(monkeypatch, msgs):
    run_mocked = RunMocked()
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs))
    monkeypatch.setattr(targetinitramfsgenerator, 'run', run_mocked)
    # FIXME
    monkeypatch.setattr(targetinitramfsgenerator, '_copy_modules', lambda *_: None)

    with pytest.raises(StopActorExecutionError) as e:
        targetinitramfsgenerator.process()
    assert 'Cannot get version of the installed RHEL-8 kernel' in str(e)
    assert not run_mocked.called


def mk_kernel_info(kernel_ver):
    kernel_info = InstalledTargetKernelInfo(pkg_nevra='nevra',
                                            kernel_img_path='vmlinuz',
                                            uname_r=kernel_ver,
                                            initramfs_path='initramfs')
    return kernel_info


@pytest.mark.parametrize('msgs', TEST_CASES)
def test_dracut_fail(monkeypatch, msgs):
    run_mocked = RunMocked(raise_err=True)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs))
    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(msgs=msgs + [mk_kernel_info(KERNEL_VERSION)]))
    monkeypatch.setattr(targetinitramfsgenerator, 'run', run_mocked)
    # FIXME
    monkeypatch.setattr(targetinitramfsgenerator, '_copy_modules', lambda *_: None)

    with pytest.raises(StopActorExecutionError) as e:
        targetinitramfsgenerator.process()
    assert 'Cannot regenerate dracut image' in str(e)
    assert run_mocked.called


@pytest.mark.parametrize(
    'msgs,files,dracut_modules,kernel_modules',
    [
        # deprecated set
        ([gen_InitrdIncludes(FILES[0])], FILES[0:1], [], []),
        ([gen_InitrdIncludes(FILES)], FILES, [], []),
        ([gen_InitrdIncludes(FILES[0:3]), gen_InitrdIncludes(FILES[3:])], FILES, [], []),
        ([gen_InitrdIncludes(FILES[0:3]), gen_InitrdIncludes(FILES)], FILES, [], []),

        # new set for files only
        ([gen_TIT([], [], FILES[0])], FILES[0:1], [], []),
        ([gen_TIT([], [], FILES)], FILES, [], []),
        ([gen_TIT([], [], FILES[0:3]), gen_TIT([], [], FILES[3:])], FILES, [], []),
        ([gen_TIT([], [], FILES[0:3]), gen_TIT([], [], FILES)], FILES, [], []),

        # deprecated and new msgs for files only
        ([gen_InitrdIncludes(FILES[0:3]), gen_TIT([], [], FILES[3:])], FILES, [], []),

        # dracut modules only
        ([gen_TIT(MODULES[0], [], [])], [], MODULES[0:1], []),
        ([gen_TIT(MODULES, [], [])], [], MODULES, []),
        ([gen_TIT(MODULES[0:3], [], []), gen_TIT(MODULES[3], [], [])], [], MODULES, []),

        # kernel modules only
        ([gen_TIT([], MODULES[0], [])], [], [], MODULES[0:1]),
        ([gen_TIT([], MODULES, [])], [], [], MODULES),
        ([gen_TIT([], MODULES[0:3], []), gen_TIT([], MODULES[3], [])], [], [], MODULES),

        # modules only - duplicates; see notes in the library
        ([gen_TIT(MODULES[0:3], [], []), gen_TIT(MODULES, [], [])], [], MODULES, []),
        ([gen_TIT([], MODULES[0:3], []), gen_TIT([], MODULES, [])], [], [], MODULES),

        # modules + files (new only)
        ([gen_TIT(MODULES, [], FILES)], FILES, MODULES, []),
        ([gen_TIT([], MODULES, FILES)], FILES, [], MODULES),

        ([gen_TIT(MODULES[0:3], [], FILES[0:3]), gen_TIT(MODULES[3:], [], FILES[3:])], FILES, MODULES, []),
        ([gen_TIT([], MODULES[0:3], FILES[0:3]), gen_TIT([], MODULES[3:], FILES[3:])], FILES, [], MODULES),

        ([gen_TIT(MODULES, [], []), gen_TIT([], [], FILES)], FILES, MODULES, []),
        ([gen_TIT([], MODULES, []), gen_TIT([], [], FILES)], FILES, [], MODULES),

        # kernel + dracut modules
        (
            [
                gen_TIT(MODULES[0:3], MODULES[0:3], FILES[0:3]),
                gen_TIT(MODULES[3:], MODULES[3:], FILES[3:])
            ],
            FILES, MODULES, MODULES
        ),

        # modules + files with deprecated msgs
        ([gen_TIT(MODULES, [], []), gen_InitrdIncludes(FILES)], FILES, MODULES, []),
        ([gen_TIT([], MODULES, []), gen_InitrdIncludes(FILES)], FILES, [], MODULES),

        ([gen_TIT(MODULES, [], FILES[0:3]), gen_InitrdIncludes(FILES[3:])], FILES, MODULES, []),
        ([gen_TIT([], MODULES, FILES[0:3]), gen_InitrdIncludes(FILES[3:])], FILES, [], MODULES),
    ])
def test_flawless(monkeypatch, msgs, files, dracut_modules, kernel_modules):
    _msgs = msgs + [mk_kernel_info(KERNEL_VERSION)]
    run_mocked = RunMocked()
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=_msgs))
    monkeypatch.setattr(targetinitramfsgenerator, 'run', run_mocked)
    # FIXME
    monkeypatch.setattr(targetinitramfsgenerator, '_copy_modules', lambda *_: None)

    targetinitramfsgenerator.process()
    assert run_mocked.called

    # check files
    if files:
        assert '--install' in run_mocked.args
        arg = run_mocked.args[run_mocked.args.index('--install') + 1]
        for f in files:
            assert f in arg
    else:
        assert '--install' not in run_mocked.args

    # check dracut modules
    if dracut_modules:
        assert '--add' in run_mocked.args
        arg = run_mocked.args[run_mocked.args.index('--add') + 1]
        for m in dracut_modules:
            assert m[0] in arg
    else:
        assert '--add' not in run_mocked.args

    # check kernel modules
    if kernel_modules:
        assert '--add-drivers' in run_mocked.args
        arg = run_mocked.args[run_mocked.args.index('--add-drivers') + 1]
        for m in kernel_modules:
            assert m[0] in arg
    else:
        assert '--add-drivers' not in run_mocked.args



import os
import shutil
import sys
import tempfile

import distro
import pytest

from leapp.libraries.actor.missinggpgkey import _expand_vars, _get_abs_file_path, _get_repo_gpgkey_urls
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import InstalledRPM, RepositoryData, RPM, TargetUserSpaceInfo


@pytest.mark.parametrize('data, exp', [
    ('bare string', 'bare string'),
    ('with dollar$$$', 'with dollar$$$'),
    ('path/with/$basearch/something', 'path/with/x86_64/something'),
    ('path/with/$releasever/something', 'path/with/9/something'),
    ('path/with/$releasever/$basearch', 'path/with/9/x86_64'),
    ('path/with/$releasever/$basearch', 'path/with/9/x86_64'),
])
def test_expand_vars(monkeypatch, data, exp):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(dst_ver='9.1'))  # x86_64 arch is default
    res = _expand_vars(data)
    assert res == exp


@pytest.mark.parametrize('repo, exp', [
    (RepositoryData(repoid='dummy', name='name'), None),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{}'), None),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgcheck":"1"}'), None),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgcheck":"0"}'), []),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgcheck":"no"}'), []),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgcheck":"False"}'), []),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgkey":"dummy"}'), ["dummy"]),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgcheck":"1","gpgkey":"dummy"}'),
     ["dummy"]),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgkey":"dummy, another"}'),
     ["dummy", "another"]),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgkey":"dummy\\nanother"}'),
     ["dummy", "another"]),
    (RepositoryData(repoid='dummy', name='name', additional_fields='{"gpgkey":"$releasever"}'),
     ["9"]),
])
def test_get_repo_gpgkey_urls(monkeypatch, repo, exp):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(dst_ver='9.1'))
    keys = _get_repo_gpgkey_urls(repo)
    assert keys == exp


@pytest.mark.parametrize('target_userspace, file_url, exists_in_container, exp', [
    (TargetUserSpaceInfo(path='/', scratch='', mounts=''), 'file:///path/to/key', True, '/path/to/key'),
    (TargetUserSpaceInfo(path='/', scratch='', mounts=''), 'file:///path/to/key', False, '/path/to/key'),
    (TargetUserSpaceInfo(path='/path/to/container/', scratch='', mounts=''), 'file:///path/to/key', True,
     '/path/to/container/path/to/key'),
    (TargetUserSpaceInfo(path='/path/to/container/', scratch='', mounts=''), 'file:///path/to/key', False,
     '/path/to/key'),
    (TargetUserSpaceInfo(path='/path/to/container/', scratch='', mounts=''), 'https://example.com/path/to/key',
     True, 'https://example.com/path/to/key'),
    (TargetUserSpaceInfo(path='/path/to/container/', scratch='', mounts=''), 'https://example.com/path/to/key',
     False, 'https://example.com/path/to/key'),
])
def test_get_abs_file_path(monkeypatch, target_userspace, file_url, exists_in_container, exp):
    def os_path_exists_mocked(path):
        if path == os.path.join(target_userspace.path, file_url[8:]) and exists_in_container:
            return True
        return False

    monkeypatch.setattr('os.path.exists', os_path_exists_mocked)
    path = _get_abs_file_path(target_userspace, file_url)
    assert path == exp



import pytest

from leapp.libraries.common.config import version
from leapp.models import FIPSInfo, Report
from leapp.utils.report import is_inhibitor


@pytest.mark.parametrize(('fips_info', 'target_major_version', 'should_inhibit'), [
    (FIPSInfo(is_enabled=True), '8', True),
    (FIPSInfo(is_enabled=True), '9', False),
    (FIPSInfo(is_enabled=False), '8', False),
    (FIPSInfo(is_enabled=False), '9', False),
])
def test_check_fips(monkeypatch, current_actor_context, fips_info, target_major_version, should_inhibit):
    monkeypatch.setattr(version, 'get_target_major_version', lambda: target_major_version)
    current_actor_context.feed(fips_info)
    current_actor_context.run()
    if should_inhibit:
        output = current_actor_context.consume(Report)
        assert len(output) == 1
        assert is_inhibitor(output[0].report)
    else:
        assert not any(is_inhibitor(msg.report) for msg in current_actor_context.consume(Report))



import pytest

from leapp.libraries.actor import adjustlocalrepos

REPO_FILE_1_LOCAL_REPOIDS = ['myrepo1']
REPO_FILE_1 = [['[myrepo1]',
                'name=mylocalrepo',
                'baseurl=file:///home/user/.local/myrepos/repo1'
                ]]
REPO_FILE_1_ADJUSTED = [['[myrepo1]',
                         'name=mylocalrepo',
                         'baseurl=file:///installroot/home/user/.local/myrepos/repo1'
                         ]]

REPO_FILE_2_LOCAL_REPOIDS = ['myrepo3']
REPO_FILE_2 = [['[myrepo2]',
                'name=mynotlocalrepo',
                'baseurl=https://www.notlocal.com/packages'
                ],
               ['[myrepo3]',
                'name=mylocalrepo',
                'baseurl=file:///home/user/.local/myrepos/repo3',
                'mirrorlist=file:///home/user/.local/mymirrors/repo3.txt'
                ]]
REPO_FILE_2_ADJUSTED = [['[myrepo2]',
                         'name=mynotlocalrepo',
                         'baseurl=https://www.notlocal.com/packages'
                         ],
                        ['[myrepo3]',
                         'name=mylocalrepo',
                         'baseurl=file:///installroot/home/user/.local/myrepos/repo3',
                         'mirrorlist=file:///installroot/home/user/.local/mymirrors/repo3.txt'
                         ]]

REPO_FILE_3_LOCAL_REPOIDS = ['myrepo4', 'myrepo5']
REPO_FILE_3 = [['[myrepo4]',
                'name=myrepowithlocalgpgkey',
                'baseurl="file:///home/user/.local/myrepos/repo4"',
                'gpgkey=file:///home/user/.local/pki/gpgkey',
                'gpgcheck=1'
                ],
               ['[myrepo5]',
                'name=myrepowithcomment',
                'baseurl=file:///home/user/.local/myrepos/repo5',
                '#baseurl=file:///home/user/.local/myotherrepos/repo5',
                'enabled=1',
                'exclude=sed']]
REPO_FILE_3_ADJUSTED = [['[myrepo4]',
                         'name=myrepowithlocalgpgkey',
                         'baseurl=file:///installroot/home/user/.local/myrepos/repo4',
                         'gpgkey=file:///home/user/.local/pki/gpgkey',
                         'gpgcheck=1'
                         ],
                        ['[myrepo5]',
                         'name=myrepowithcomment',
                         'baseurl=file:///installroot/home/user/.local/myrepos/repo5',
                         '#baseurl=file:///home/user/.local/myotherrepos/repo5',
                         'enabled=1',
                         'exclude=sed']]
REPO_FILE_EMPTY = []


@pytest.mark.parametrize('repo_file_line, expected_adjusted_repo_file_line',
                         [('baseurl=file:///home/user/.local/repositories/repository',
                           'baseurl=file:///installroot/home/user/.local/repositories/repository'),
                          ('baseurl="file:///home/user/my-repo"',
                           'baseurl=file:///installroot/home/user/my-repo'),
                          ('baseurl=https://notlocal.com/packages',
                           'baseurl=https://notlocal.com/packages'),
                          ('mirrorlist=file:///some_mirror_list.txt',
                           'mirrorlist=file:///installroot/some_mirror_list.txt'),
                          ('gpgkey=file:///etc/pki/some.key',
                           'gpgkey=file:///etc/pki/some.key'),
                          ('#baseurl=file:///home/user/my-repo',
                           '#baseurl=file:///home/user/my-repo'),
                          ('', ''),
                          ('[repoid]', '[repoid]')])
def test_adjust_local_file_url(repo_file_line, expected_adjusted_repo_file_line):
    adjusted_repo_file_line = adjustlocalrepos._adjust_local_file_url(repo_file_line)
    if 'file://' not in repo_file_line:
        assert adjusted_repo_file_line == repo_file_line
        return
    assert adjusted_repo_file_line == expected_adjusted_repo_file_line


class MockedFileDescriptor(object):

    def __init__(self, repo_file, expected_new_repo_file):
        self.repo_file = repo_file
        self.expected_new_repo_file = expected_new_repo_file

    @staticmethod
    def _create_repo_file_lines(repo_file):
        repo_file_lines = []
        for repo in repo_file:
            repo = [line+'\n' for line in repo]
            repo_file_lines += repo
        return repo_file_lines

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        return

    def readlines(self):
        return self._create_repo_file_lines(self.repo_file)

    def write(self, new_contents):
        assert self.expected_new_repo_file
        repo_file_lines = self._create_repo_file_lines(self.expected_new_repo_file)
        expected_repo_file_contents = ''.join(repo_file_lines).rstrip('\n')
        assert expected_repo_file_contents == new_contents


class MockedContext(object):

    def __init__(self, repo_contents, expected_repo_contents):
        self.repo_contents = repo_contents
        self.expected_repo_contents = expected_repo_contents

    def open(self, path, mode):
        return MockedFileDescriptor(self.repo_contents, self.expected_repo_contents)


@pytest.mark.parametrize('repo_file, local_repoids, expected_repo_file',
                         [(REPO_FILE_1, REPO_FILE_1_LOCAL_REPOIDS, REPO_FILE_1_ADJUSTED),
                          (REPO_FILE_2, REPO_FILE_2_LOCAL_REPOIDS, REPO_FILE_2_ADJUSTED),
                          (REPO_FILE_3, REPO_FILE_3_LOCAL_REPOIDS, REPO_FILE_3_ADJUSTED)])
def test_adjust_local_repos_to_container(repo_file, local_repoids, expected_repo_file):
    # The checks for expected_repo_file comparison to a adjusted form of the
    # repo_file can be found in the MockedFileDescriptor.write().
    context = MockedContext(repo_file, expected_repo_file)
    adjustlocalrepos._adjust_local_repos_to_container(context, '<some_repo_file_path>', local_repoids)


@pytest.mark.parametrize('expected_repo_file, add_empty_lines', [(REPO_FILE_EMPTY, False),
                                                                 (REPO_FILE_1, False),
                                                                 (REPO_FILE_2, True)])
def test_extract_repos_from_repofile(expected_repo_file, add_empty_lines):
    repo_file = expected_repo_file[:]
    if add_empty_lines:  # add empty lines before the first repo
        repo_file[0] = ['', ''] + repo_file[0]

    context = MockedContext(repo_file, None)
    repo_gen = adjustlocalrepos._extract_repos_from_repofile(context, '<some_repo_file_path>')

    for repo in expected_repo_file:
        assert repo == next(repo_gen, None)

    assert next(repo_gen, None) is None



from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api


def process(openssh_messages):
    removed_ciphers = [
        "blowfish-cbc",
        "cast128-cbc",
        "arcfour",
        "arcfour128",
        "arcfour256",
    ]
    removed_macs = [
        "hmac-ripemd160",
    ]
    found_ciphers = []
    found_macs = []

    config = next(openssh_messages, None)
    if list(openssh_messages):
        api.current_logger().warning('Unexpectedly received more than one OpenSshConfig message.')
    if not config:
        raise StopActorExecutionError(
            'Could not check openssh configuration', details={'details': 'No OpenSshConfig facts found.'}
        )

    for cipher in removed_ciphers:
        if config.ciphers and cipher in config.ciphers:
            found_ciphers.append(cipher)
    for mac in removed_macs:
        if config.macs and mac in config.macs:
            found_macs.append(mac)

    resources = [
        reporting.RelatedResource('package', 'openssh-server'),
        reporting.RelatedResource('file', '/etc/ssh/sshd_config')
    ]
    if found_ciphers:
        reporting.create_report([
            reporting.Title('OpenSSH configured to use removed ciphers'),
            reporting.Summary(
                'OpenSSH is configured to use removed ciphers {}. '
                'These ciphers were removed from OpenSSH and if '
                'present the sshd daemon will not start in RHEL 8'
                ''.format(','.join(found_ciphers))
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
            reporting.Remediation(
                hint='Remove the following ciphers from sshd_config: '
                     '{}'.format(','.join(found_ciphers))
            ),
            reporting.Groups([reporting.Groups.INHIBITOR])
        ] + resources)

    if found_macs:
        reporting.create_report([
            reporting.Title('OpenSSH configured to use removed mac'),
            reporting.Summary(
                'OpenSSH is configured to use removed mac {}. '
                'This MAC was removed from OpenSSH and if present '
                'the sshd daemon will not start in RHEL 8'
                ''.format(','.join(found_macs))
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
            reporting.Remediation(
                hint='Remove the following MACs from sshd_config: {}'.format(','.join(found_macs))
            ),
            reporting.Groups([reporting.Groups.INHIBITOR])
        ] + resources)



from leapp.actors import Actor
from leapp.libraries.actor import multipathconfcheck
from leapp.models import MultipathConfFacts
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class MultipathConfCheck(Actor):
    """
    Checks whether the multipath configuration can be updated to RHEL-8 and
    plan necessary tasks.

    Specifically, it checks if the path_checker/checker option is set to
    something other than tur in the defaults section. If so, non-trivial
    changes may be required in the multipath.conf file, and it is not
    possible to auto-update it - in such a case inhibit upgrade.

    In addition create a task to ensure that configuration files are copied
    into the target container (they are necessary for correct creation of the
    upgrade initramfs.
    """

    name = 'multipath_conf_check'
    consumes = (MultipathConfFacts,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        facts = next(self.consume(MultipathConfFacts), None)
        if facts is None:
            self.log.debug('Skipping execution. No MultipathConfFacts has '
                           'been produced')
            return
        multipathconfcheck.check_configs(facts)



from leapp.actors import Actor
from leapp.libraries.actor import checkinstalleddebugkernels
from leapp.models import DistributionSignedRPM
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckInstalledDebugKernels(Actor):
    """
    Inhibit IPU (in-place upgrade) when multiple debug kernels are installed.

    Because of an issue in DNF, the transaction can't be validated if there's
    more than one package named kernel-debug. Therefore, in this case, we
    inhibit the upgrade with a clearer remediation.
    """

    name = 'check_installed_debug_kernels'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        checkinstalleddebugkernels.process()



from leapp.actors import Actor
from leapp.libraries.actor import targetuserspacecryptopolicies
from leapp.models import CryptoPolicyInfo, TargetUserSpaceInfo
from leapp.tags import IPUWorkflowTag, TargetTransactionChecksPhaseTag

# TODO(pstodulk): move to common repo in future. This is something what should
# be in future generalized for all IPUs. But right now, we have blocking
# issues just for 8 -> 9.


class TargetUserspaceCryptoPolicies(Actor):
    """
    Set crypto policies inside the target userspace container.

    The crypto policies inside the container are "DEFAULT" by default, ignoring
    the setup on the host system. This leads to situations like when
    we work with rpms differently than the host system after the upgrade.
    The policies inside the container should reflect policies on the host
    system from the point of the target OS. E.g. when FIPS is used, we should
    not use DEFAULT, or when LEGACY is used, we should use the LEGACY as well.

    However, right now we are not able to handle systems with the custom crypto
    policies. In such a case the actor raises an error and upgrade is stopped.
    """

    name = 'target_userspace_crypto_policies'
    consumes = (CryptoPolicyInfo, TargetUserSpaceInfo)
    produces = ()
    tags = (IPUWorkflowTag, TargetTransactionChecksPhaseTag)

    def process(self):
        targetuserspacecryptopolicies.process()



# plugin inspired by "system_upgrade.py" from rpm-software-management
from __future__ import print_function

import json
import logging
import sys

import dnf
import dnf.cli
import dnf.module.module_base

CMDS = ['check', 'download', 'dry-run', 'upgrade']
"""
Basic subcommands for the plugin.

    check    -> check we are able to calculate the DNF transaction using
                only the YUM/DNF repositories metadata (no packages downloaded)
    download -> calculate the transaction, download all packages and test
                the transaction is duable using data contained in RPMs
                (include search for file conflicts, disk space, ...)
    dry-run  -> test the transaction again with the cached content (including
                the downloaded RPMs; it is subset of the download cmd)(
                  BZ: 1976932 - not enough space on disk,
                  PR: https://github.com/oamg/leapp-repository/pull/734
                )
    upgrade  -> perform the DNF transaction using the cached data only
"""


class DoNotDownload(Exception):
    pass


def _do_not_download_packages(packages, progress=None, total=None):
    raise DoNotDownload()


class RhelUpgradeCommand(dnf.cli.Command):
    aliases = ('rhel-upgrade',)
    summary = 'Plugin for upgrading to the next RHEL major release'

    def __init__(self, cli):
        super(RhelUpgradeCommand, self).__init__(cli)
        self.plugin_data = {}

    @staticmethod
    def set_argparser(parser):
        parser.add_argument('tid', nargs=1, choices=CMDS,
                            metavar="[%s]" % "|".join(CMDS))
        parser.add_argument('filename')

    def _process_entities(self, entities, op, entity_name):
        """
        Adds list of packages for given operation to the transaction
        """
        entities_notfound = []

        for spec in entities:
            try:
                op(spec)
            except dnf.exceptions.MarkingError:
                if isinstance(spec, (list, tuple)):
                    entities_notfound.extend(spec)
                else:
                    entities_notfound.append(spec)
        if entities_notfound:
            err_str = ('{} marked by Leapp to {} not found '
                       'in repositories metadata: '.format(entity_name, op.__name__) + ' '.join(entities_notfound))
            print('Warning: ' + err_str, file=sys.stderr)

    def _save_aws_region(self, region):
        self.plugin_data['rhui']['aws']['region'] = region
        with open(self.opts.filename, 'w+') as fo:
            json.dump(self.plugin_data, fo, sort_keys=True, indent=2)

    def _read_aws_region(self, repo):
        region = None
        if repo.baseurl:
            # baseurl is tuple (changed by Amazon-id plugin)
            # here we take just the first baseurl as the REGION will be same for all of them
            region = repo.baseurl[0].split('.', 2)[1]
        elif repo.mirrorlist:
            region = repo.mirrorlist.split('.', 2)[1]
        if not region:
            print('Could not read AWS REGION from either baseurl or mirrorlist', file=sys.stderr)
            sys.exit(1)
        return region

    def _fix_rhui_url(self, repo, region):
        if repo.baseurl:
            repo.baseurl = tuple(
                url.replace('REGION', region, 1) for url in repo.baseurl
            )
        elif repo.mirrorlist:
            repo.mirrorlist = repo.mirrorlist.replace('REGION', region, 1)
        else:
            raise dnf.exceptions.RepoError("RHUI repository %s does not have an url" % repo.name)
        return repo

    def pre_configure(self):
        with open(self.opts.filename) as fo:
            self.plugin_data = json.load(fo)
        # There is an issue that ignores releasever value if it is set at configure
        self.base.conf.releasever = self.plugin_data['dnf_conf']['releasever']

    def configure(self):

        on_aws = self.plugin_data['rhui']['aws']['on_aws']
        self.cli.demands.root_user = True
        self.cli.demands.resolving = self.opts.tid[0] != 'check'
        self.cli.demands.available_repos = True
        self.cli.demands.sack_activation = True
        self.cli.demands.cacheonly = self.opts.tid[0] in ['dry-run', 'upgrade']
        self.cli.demands.allow_erasing = self.plugin_data['dnf_conf']['allow_erasing']
        self.base.conf.protected_packages = []
        self.base.conf.best = self.plugin_data['dnf_conf']['best']
        self.base.conf.assumeyes = True
        self.base.conf.gpgcheck = self.plugin_data['dnf_conf']['gpgcheck']
        self.base.conf.debug_solver = self.plugin_data['dnf_conf']['debugsolver']
        self.base.conf.module_platform_id = self.plugin_data['dnf_conf']['platform_id']
        installroot = self.plugin_data['dnf_conf'].get('installroot')
        if installroot:
            self.base.conf.installroot = installroot
        if self.plugin_data['dnf_conf']['test_flag'] and self.opts.tid[0] in ['download', 'dry-run']:
            self.base.conf.tsflags.append("test")

        enabled_repos = self.plugin_data['dnf_conf']['enable_repos']
        self.base.repos.all().disable()

        aws_region = None

        for repo in self.base.repos.all():
            if repo.id in enabled_repos:
                repo.skip_if_unavailable = False
                if not self.base.conf.gpgcheck:
                    repo.gpgcheck = False
                repo.enable()
                if self.opts.tid[0] == 'download' and on_aws:
                    # during the upgrade phase we has to disable "Amazon-id" plugin as we do not have networking
                    # in initramdisk (yet, but we probably do not want it to do anything anyway as we already have
                    # packages downloaded and cached). However, when we disable it, the plugin cannot substitute
                    # "REGION" placeholder in mirrorlist url and consequently we cannot identify a correct cache
                    # folder in "/var/cache/dnf" as it has different digest calculated based on already substituted
                    # placeholder.
                    # E.g
                    # "https://rhui3.REGION.aws.ce.redhat.com" becomes "https://rhui3.eu-central-1.aws.ce.redhat.com"
                    #
                    # region should be same for all repos so we are fine to collect it from
                    # the last one
                    aws_region = self._read_aws_region(repo)
                if self.opts.tid[0] in ['dry-run', 'upgrade'] and on_aws:
                    aws_region = self.plugin_data['rhui']['aws']['region']
                    if aws_region:
                        repo = self._fix_rhui_url(repo, aws_region)

        if aws_region and self.opts.tid[0] == 'download':
            self._save_aws_region(aws_region)

    def run(self):
        # takes local rpms, creates Package objects from them, and then adds them to the sack as virtual repository
        local_rpm_objects = self.base.add_remote_rpms(self.plugin_data['pkgs_info']['local_rpms'])

        for pkg in local_rpm_objects:
            self.base.package_install(pkg)

        module_base = dnf.module.module_base.ModuleBase(self.base)

        # Module tasks
        modules_to_enable = self.plugin_data['pkgs_info'].get('modules_to_enable', ())

        available_modules_to_enable = []
        unavailable_modules = []
        for module in modules_to_enable:
            matching_modules, dummy_nsvcap = module_base.get_modules(module)
            target_bucket = available_modules_to_enable if matching_modules else unavailable_modules
            target_bucket.append(module)

        if unavailable_modules:
            dnf_plugin_logger = logging.getLogger('dnf.plugin')
            msg = 'The following modules were requested to be enabled, but they are unavailable: %s'
            dnf_plugin_logger.warning(msg, ', '.join(unavailable_modules))

        # Package tasks
        to_install = self.plugin_data['pkgs_info']['to_install']
        to_remove = self.plugin_data['pkgs_info']['to_remove']
        to_upgrade = self.plugin_data['pkgs_info']['to_upgrade']

        # Modules to enable
        self._process_entities(entities=[available_modules_to_enable],
                               op=module_base.enable,
                               entity_name='Module stream')

        # Packages to be removed
        self._process_entities(entities=to_remove, op=self.base.remove, entity_name='Package')
        # Packages to be installed
        self._process_entities(entities=to_install, op=self.base.install, entity_name='Package')
        # Packages to be upgraded
        self._process_entities(entities=to_upgrade, op=self.base.upgrade, entity_name='Package')
        self.base.distro_sync()

        if self.opts.tid[0] == 'check':
            try:
                self.base.resolve(allow_erasing=self.cli.demands.allow_erasing)
            except dnf.exceptions.DepsolveError as e:
                print('Transaction check: ', file=sys.stderr)
                print(str(e), file=sys.stderr)
                raise

            # We are doing this to avoid downloading the packages in the check phase
            self.base.download_packages = _do_not_download_packages
            try:
                displays = []
                if self.cli.demands.transaction_display is not None:
                    displays.append(self.cli.demands.transaction_display)
                self.base.do_transaction(display=displays)
            except DoNotDownload:
                print('Check completed.')


class RhelUpgradePlugin(dnf.Plugin):
    name = 'rhel-upgrade'

    def __init__(self, base, cli):
        super(RhelUpgradePlugin, self).__init__(base, cli)
        if cli:
            cli.register_command(RhelUpgradeCommand)



from leapp.actors import Actor
from leapp.libraries.actor import multipathconfupdate
from leapp.models import MultipathConfFacts8to9
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MultipathConfUpdate8to9(Actor):
    """
    Modifies multipath configuration files on the target RHEL-9 system so that
    they will run properly. This is done in three ways
    1. Adding the allow_usb_devices and enable_foreign options to
       /etc/multipath.conf if they are not present, to retain RHEL-8 behavior
    2. Converting any "*" regular expression strings to ".*"
    """

    name = 'multipath_conf_update_8to9'
    consumes = (MultipathConfFacts8to9,)
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        facts = next(self.consume(MultipathConfFacts8to9), None)
        if facts is None:
            self.log.debug('Skipping execution. No MultipathConfFacts8to9 has '
                           'been produced')
            return
        multipathconfupdate.update_configs(facts)



from leapp.actors import Actor
from leapp.libraries.actor import create_iso_repofile
from leapp.models import CustomTargetRepositoryFile, TargetOSInstallationImage
from leapp.tags import IPUWorkflowTag, TargetTransactionFactsPhaseTag


class CreateISORepofile(Actor):
    """
    Create custom repofile containing information about repositories found in target OS installation ISO, if used.
    """

    name = 'create_iso_repofile'
    consumes = (TargetOSInstallationImage,)
    produces = (CustomTargetRepositoryFile,)
    tags = (IPUWorkflowTag, TargetTransactionFactsPhaseTag)

    def process(self):
        create_iso_repofile.produce_repofile_if_iso_used()



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import isccfg
from leapp.libraries.stdlib import api, CalledProcessError, run

# Callback for walk function
callbacks = {
    'dnssec-lookaside': isccfg.ModifyState.callback_comment_out,
}


def paths_from_issues(issues):
    """Extract paths from list of BindConfigIssuesModel."""
    return [issue.path for issue in issues]


def parser_file(parser, path):
    for cfg in parser.FILES_TO_CHECK:
        if cfg.path == path:
            return cfg
    return None


def make_backup(path, backup_suffix='.leapp'):
    """Make backup of a file before modification."""
    backup_path = path + backup_suffix
    try:
        run(['cp', '--preserve=all', path, backup_path])
    except CalledProcessError as exc:
        raise StopActorExecutionError(
            'Could not create a backup copy',
            details={'details': 'An exception during backup raised {}'.format(str(exc))}
        )


def update_section(parser, section):
    """Modify one section.

    :type section: ConfigSection
    """
    state = isccfg.ModifyState()
    parser.walk(section, callbacks, state)
    state.finish(section)
    return state.content()


def update_config(parser, cfg):
    """Modify contents of file according to rules.

    :type cfg: ConfigFile
    :returns str: Modified config contents
    """
    return update_section(parser, cfg.root_section())


def update_file(parser, path, write=True):
    """Prepare modified content for the file, make backup and rewrite it.

    :param parser: IscConfigParser
    :param path: String with path to a file
    :param log: Log instance with debug(str) method or None
    :param write: True to allow file modification, false to only return modification status
    """
    cfg = parser_file(parser, path)
    modified = update_config(parser, cfg)
    if modified != cfg.buffer:
        api.current_logger().debug('%s needs modification', path)
        if write:
            make_backup(path)
            with open(path, 'w') as f:
                f.write(modified)
            api.current_logger().debug('%s updated to size %d', path, len(modified))
        return True
    return False


def update_facts(facts, path='/etc/named.conf'):
    """Parse and update all files according to supplied facts.

    :param facts: BindFacts instance
    :param path: String to main configuration file
    :returns: number of modified files
    """
    parser = isccfg.IscConfigParser(path)
    modified_files = set()
    if facts.dnssec_lookaside:
        for model in facts.dnssec_lookaside:
            if update_file(parser, model.path):
                modified_files.add(model.path)
    facts.modified_files = list(modified_files)



from leapp import reporting
from leapp.libraries.common import rhsm
from leapp.libraries.common.config import get_env
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import InstalledRPM, RpmTransactionTasks

INSIGHTS_CLIENT_PKG = "insights-client"


def _ensure_package(package):
    """
    Produce install tasks if the given package is missing

    :return: True if the install task is produced else False
    """
    has_client_package = has_package(InstalledRPM, package)
    if not has_client_package:
        api.produce(RpmTransactionTasks(to_install=[package]))

    return not has_client_package


def _report_registration_info(installing_client):
    pkg_msg = " The '{}' package required for the registration will be installed during the upgrade."

    title = "Automatic registration into Red Hat Insights"
    summary = (
        "After the upgrade, this system will be automatically registered into Red Hat Insights."
        "{}"
        " To skip the automatic registration, use the '--no-insights-register' command line option or"
        " set the LEAPP_NO_INSIGHTS_REGISTER environment variable."
    ).format(pkg_msg.format(INSIGHTS_CLIENT_PKG) if installing_client else "")

    reporting.create_report(
        [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.INFO),
            reporting.Groups([reporting.Groups.SERVICES]),
        ]
    )


def process():
    if rhsm.skip_rhsm():
        return

    if get_env("LEAPP_NO_INSIGHTS_REGISTER", "0") == "1":
        return

    installing_client = _ensure_package(INSIGHTS_CLIENT_PKG)
    _report_registration_info(installing_client)



import os
from xml.etree import ElementTree

from leapp.actors import Actor
from leapp.libraries.actor import firewalldfactsactor
from leapp.models import FirewalldFacts
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class FirewalldFactsActor(Actor):
    """
    Provide data about firewalld

    After collecting data, a message with relevant data will be produced.
    """

    name = 'firewalld_facts_actor'
    consumes = ()
    produces = (FirewalldFacts,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        facts = FirewalldFacts()

        try:
            tree = ElementTree.parse('/etc/firewalld/lockdown-whitelist.xml')
            root = tree.getroot()
            facts.firewall_config_command = firewalldfactsactor.getLockdownFirewallConfigCommand(root)
        except IOError:
            pass

        try:
            tree = ElementTree.parse('/etc/firewalld/direct.xml')
            root = tree.getroot()
            facts.ebtablesTablesInUse = firewalldfactsactor.getEbtablesTablesInUse(root)
        except IOError:
            pass

        ipsetTypesInUse = set()
        directory = '/etc/firewalld/ipsets'
        try:
            for filename in os.listdir(directory):
                if not filename.endswith('.xml'):
                    continue
                try:
                    tree = ElementTree.parse(os.path.join(directory, filename))
                    root = tree.getroot()
                    ipsetTypesInUse |= set(firewalldfactsactor.getIpsetTypesInUse(root))
                except IOError:
                    pass
            facts.ipsetTypesInUse = list(ipsetTypesInUse)
        except OSError:
            pass

        self.produce(facts)



import pytest

from leapp.libraries.actor import insightsautoregister
from leapp.libraries.common import rhsm
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import api, CalledProcessError


@pytest.mark.parametrize(
    ('skip_rhsm', 'no_register', 'should_register'),
    [
        (False, False, True),
        (False, True, False),
        (True, False, False),
        (True, True, False),
    ]
)
def test_should_register(monkeypatch, skip_rhsm, no_register, should_register):
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: skip_rhsm)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        envars={'LEAPP_NO_INSIGHTS_REGISTER': '1' if no_register else '0'}
        ))

    called = []

    def _insights_register_mocked():
        called.append(True)

    monkeypatch.setattr(
        insightsautoregister,
        '_insights_register',
        _insights_register_mocked
    )

    insightsautoregister.process()

    assert len(called) == should_register


def test_insights_register_success_logged(monkeypatch):

    def run_mocked(cmd, **kwargs):
        return {
            'stdout': 'Successfully registered into Insights',
            'stderr': '',
            'exit_code': 0
        }

    monkeypatch.setattr(insightsautoregister, 'run', run_mocked)
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    insightsautoregister._insights_register()

    assert api.current_logger.infomsg
    assert not api.current_logger.errmsg


def test_insights_register_failure_logged(monkeypatch):

    def run_mocked(cmd, **kwargs):
        raise CalledProcessError(
            message='A Leapp Command Error occurred.',
            command=cmd,
            result={'signal': None, 'exit_code': 1, 'pid': 0, 'stdout': 'fake', 'stderr': 'fake'}
        )

    monkeypatch.setattr(insightsautoregister, 'run', run_mocked)
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    insightsautoregister._insights_register()

    assert not api.current_logger.infomsg
    assert api.current_logger.errmsg



from leapp.actors import Actor
from leapp.libraries.actor import migrateblacklistca
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MigrateBlacklistCA(Actor):
    """
    Preserve blacklisted certificates during the upgrade

    Path for the blacklisted certificates has been changed on RHEL 9.
    The original paths on RHEL 8 and older systems have been:
        /etc/pki/ca-trust/source/blacklist/
        /usr/share/pki/ca-trust-source/blacklist/
    However on RHEL 9 the blacklist directory has been renamed to 'blocklist'.
    So the new paths are:
        /etc/pki/ca-trust/source/blocklist/
        /usr/share/pki/ca-trust-source/blocklist/
    This actor moves all blacklisted certificates into the expected directories
    and fix symlinks if needed.
    """

    name = 'migrate_blacklist_ca'
    consumes = ()
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        migrateblacklistca.process()



from leapp.actors import Actor
from leapp.libraries.actor import networkdeprecations
from leapp.models import IfCfg, NetworkManagerConnection, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckNetworkDeprecations(Actor):
    """
    Ensures that network configuration doesn't rely on unsupported settings

    Inhibits upgrade if the network configuration would not work with
    NetworkManager on the upgraded system due functionality being deprecated.

    Includes check for insecure wireless network encryption configuration
    what will not work with RHEL9.
    """

    name = "network_deprecations"
    consumes = (IfCfg, NetworkManagerConnection,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag,)

    def process(self):
        networkdeprecations.process()



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class RoceDetected(Model):
    """
    The model creates a list of
     - RoCE NICs that are detected as connected, which means, they are
       configured persistently
     - RoCE NICs that are in process of connecting (i.e. they are trying
       to get an IP address - and might become connected upon success)
    """
    topic = SystemInfoTopic

    roce_nics_connected = fields.List(fields.String(), default=[])
    """
    List of RoCE NICs which are detected as connected.

    e.g. ["ens1234", "eno3456"]
    """

    roce_nics_connecting = fields.List(fields.String(), default=[])
    """
    List of RoCE NICs which are detected as connecting right now.

    (They might become detected as connected, soon.)
    """



from leapp.actors import Actor
from leapp.libraries.actor.ipascanner import is_ipa_client_configured, is_ipa_server_configured
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM, IpaInfo
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class IpaScanner(Actor):
    """
    Scan system for ipa-client and ipa-server status
    """

    name = "ipa_scanner"
    consumes = (DistributionSignedRPM,)
    produces = (IpaInfo,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        ipainfo = IpaInfo(
            has_client_package=has_package(
                DistributionSignedRPM, "ipa-client"
            ),
            is_client_configured=is_ipa_client_configured(),
            has_server_package=has_package(
                DistributionSignedRPM, "ipa-server"
            ),
            is_server_configured=is_ipa_server_configured(),
        )
        self.produce(ipainfo)



from leapp.actors import Actor
from leapp.libraries.actor.authselectscanner import (
    Authconfig,
    AuthselectScannerLibrary,
    DConf,
    read_file
)
from leapp.libraries.common.pam import PAM
from leapp.models import Authselect
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class AuthselectScanner(Actor):
    """
    Detect what authselect configuration should be suggested to administrator.

    1. Detect possible authselect profile by looking up modules in PAM
       or by checking that daemon is enabled.
       - pam_sss -> sssd
       - pam_winbind -> winbind
       - ypbind enabled -> nis

       If more then one module/daemon is detected that we will keep the
       configuration intact. No authselect profile can be applied.

    2. Detect authselect profile features by looking up modules in PAM
       or nsswitch.conf.
       - pam_faillock => with-faillock
       - pam_fprintd => with-fingerprint
       - pam_access => with-pamaccess
       - pam_mkhomedir => with-mkhomedir
       - pam_oddjob_mkhomedir => with-mkhomedir

    3. Check if there are any unknown PAM modules.
       If there are used PAM modules not used in authselect (such as pam_ldap),
       we must keep the configuration intact.

    4. Check if authconfig was used to create current configuration.
       If yes, we can automatically convert the configuration to authselect.
       If no, we need admin's confirmation.

       - Check that /etc/sysconfig/authconfig exists.
       - Check that PAM configuration uses authconfig files.
       - Check that PAM configuration was not touch after sysconfig file
         was created.
    """

    name = 'authselect_scanner'
    consumes = ()
    produces = (Authselect,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    known_modules = [
        'pam_access',
        'pam_deny',
        'pam_ecryptfs',
        'pam_env',
        'pam_faildelay',
        'pam_faillock',
        'pam_fprintd',
        'pam_keyinit',
        'pam_krb5',
        'pam_lastlog',
        'pam_limits',
        'pam_localuser',
        'pam_mkhomedir',
        'pam_oddjob_mkhomedir',
        'pam_permit',
        'pam_pkcs11',
        'pam_pwquality',
        'pam_sss',
        'pam_succeed_if',
        'pam_systemd',
        'pam_u2f',
        'pam_umask',
        'pam_unix',
        'pam_winbind'
    ]
    """
    List of PAM modules that are known by authselect.
    """

    def process(self):
        # Load configuration
        ac = Authconfig(read_file('/etc/sysconfig/authconfig'))
        dconf = DConf(read_file('/etc/dconf/db/distro.d/10-authconfig'))
        pam = PAM.from_system_configuration()
        nsswitch = read_file("/etc/nsswitch.conf")

        scanner = AuthselectScannerLibrary(
            self.known_modules,
            ac, dconf, pam, nsswitch
        )

        self.produce(scanner.process())



from leapp.actors import Actor
from leapp.libraries.actor.scanfilesfortargetuserspace import scan_files_to_copy
from leapp.models import TargetUserSpacePreupgradeTasks
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanFilesForTargetUserspace(Actor):
    """
    Scan the source system and identify files that will be copied into the target userspace when it is created.
    """

    name = 'scan_files_for_target_userspace'
    consumes = ()
    produces = (TargetUserSpacePreupgradeTasks,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        scan_files_to_copy()



from leapp.actors import Actor
from leapp.libraries.actor.checkyumpluginsenabled import check_required_yum_plugins_enabled
from leapp.models import PkgManagerInfo
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckYumPluginsEnabled(Actor):
    """
    Checks that the required yum plugins are enabled.
    """

    name = 'check_yum_plugins_enabled'
    consumes = (PkgManagerInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        pkg_manager_info = next(self.consume(PkgManagerInfo))
        check_required_yum_plugins_enabled(pkg_manager_info)



import textwrap

from six import StringIO

from leapp.libraries.actor.sssdfacts8to9 import SSSDFactsLibrary
from leapp.libraries.common import utils


def get_config(content):
    return utils.parse_config(StringIO(textwrap.dedent(content)))


def test_empty_config():
    config = utils.parse_config()
    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_enable_files_domain_set__notset():
    config = get_config("""
    [sssd]
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_enable_files_domain_set__set_true():
    config = get_config("""
    [sssd]
    enable_files_domain = true
    """)

    facts = SSSDFactsLibrary(config).process()

    assert facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_enable_files_domain_set__set_false():
    config = get_config("""
    [sssd]
    enable_files_domain = false
    """)

    facts = SSSDFactsLibrary(config).process()

    assert facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_enable_files_domain_set__nodomain():
    config = get_config("""
    [sssd]
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_explicit_files_domain__notset():
    config = get_config("""
    [sssd]
    domains = ldap

    [domain/ldap]
    id_provider = ldap
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_explicit_files_domain__set():
    config = get_config("""
    [sssd]
    domains = files

    [domain/files]
    id_provider = files
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_pam_cert_auth__notset():
    config = get_config("""
    [pam]
    pam_gssapi_services = sudo
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_pam_cert_auth__true():
    config = get_config("""
    [pam]
    pam_gssapi_services = sudo
    pam_cert_auth = true
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert facts.pam_cert_auth


def test_pam_cert_auth__false():
    config = get_config("""
    [pam]
    pam_gssapi_services = sudo
    pam_cert_auth = false
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert not facts.explicit_files_domain
    assert not facts.pam_cert_auth


def test_complex_config():
    config = get_config("""
    [sssd]
    user = root
    services = nss, pam, sudo
    domain = ldap, files

    [pam]
    pam_cert_auth = true

    [domain/ldap]
    id_provider = ldap
    ldap_groups_use_matching_rule_in_chain = True

    [domain/files]
    id_provider = files
    """)

    facts = SSSDFactsLibrary(config).process()

    assert not facts.enable_files_domain_set
    assert facts.explicit_files_domain
    assert facts.pam_cert_auth



import os
from collections import namedtuple

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import kernel as kernel_lib
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import InstalledTargetKernelInfo, InstalledTargetKernelVersion, KernelInfo
from leapp.utils.deprecation import suppress_deprecation

KernelBootFiles = namedtuple('KernelBootFiles', ('vmlinuz_path', 'initramfs_path'))


def get_kernel_pkg_name(rhel_major_version, kernel_type):
    """
    Get the name of the package providing kernel binaries.

    :param str rhel_major_version: RHEL major version
    :param KernelType kernel_type: Type of the kernel
    :returns: Kernel package name
    :rtype: str
    """
    if rhel_major_version == '7':
        kernel_pkg_name_table = {
            kernel_lib.KernelType.ORDINARY: 'kernel',
            kernel_lib.KernelType.REALTIME: 'kernel-rt'
        }
    else:
        kernel_pkg_name_table = {
            kernel_lib.KernelType.ORDINARY: 'kernel-core',
            kernel_lib.KernelType.REALTIME: 'kernel-rt-core'
        }
    return kernel_pkg_name_table[kernel_type]


def get_target_kernel_package_nevra(kernel_pkg_name):
    try:
        kernel_nevras = run(['rpm', '-q', kernel_pkg_name], split=True)['stdout']
    except CalledProcessError:
        return ''

    target_kernel_el = 'el{}'.format(get_target_major_version())
    for kernel_nevra in kernel_nevras:
        if target_kernel_el in kernel_nevra:
            return kernel_nevra
    return ''


def get_boot_files_provided_by_kernel_pkg(kernel_nevra):
    initramfs_path = ''
    vmlinuz_path = ''
    err_msg = 'Cannot determine location of the target kernel boot image and corresponding initramfs .'
    try:
        kernel_pkg_files = run(['rpm', '-q', '-l', kernel_nevra], split=True)['stdout']
        for kernel_file_path in kernel_pkg_files:
            dirname = os.path.dirname(kernel_file_path)
            if dirname != '/boot':
                continue
            basename = os.path.basename(kernel_file_path)
            if basename.startswith('vmlinuz'):
                vmlinuz_path = kernel_file_path
            elif basename.startswith('initramfs'):
                initramfs_path = kernel_file_path
    except CalledProcessError:
        raise StopActorExecutionError(err_msg)
    if not vmlinuz_path or not initramfs_path:
        raise StopActorExecutionError(err_msg)
    return KernelBootFiles(vmlinuz_path=vmlinuz_path, initramfs_path=initramfs_path)


@suppress_deprecation(InstalledTargetKernelVersion)
def process():
    # pylint: disable=no-else-return  - false positive
    # TODO: should we take care about stuff of kernel-rt and kernel in the same
    # time when both are present? or just one? currently, handle only one
    # of these during the upgrade. kernel-rt has higher prio when original sys
    # was realtime
    src_kernel_info = next(api.consume(KernelInfo), None)
    if not src_kernel_info:
        return  # Will not happen, other actors would inhibit the upgrade

    target_ver = get_target_major_version()
    target_kernel_pkg_name = get_kernel_pkg_name(target_ver, src_kernel_info.type)
    target_kernel_nevra = get_target_kernel_package_nevra(target_kernel_pkg_name)

    if src_kernel_info.type != kernel_lib.KernelType.ORDINARY and not target_kernel_nevra:
        api.current_logger().warning('The kernel-rt-core rpm from the target RHEL has not been detected. Switching '
                                     'to non-preemptive kernel.')
        target_kernel_pkg_name = get_kernel_pkg_name(target_ver, kernel_lib.KernelType.ORDINARY)
        target_kernel_nevra = get_target_kernel_package_nevra(target_kernel_pkg_name)

    if target_kernel_nevra:
        boot_files = get_boot_files_provided_by_kernel_pkg(target_kernel_nevra)
        target_kernel_version = kernel_lib.get_uname_r_provided_by_kernel_pkg(target_kernel_nevra)
        installed_kernel_info = InstalledTargetKernelInfo(pkg_nevra=target_kernel_nevra,
                                                          uname_r=target_kernel_version,
                                                          kernel_img_path=boot_files.vmlinuz_path,
                                                          initramfs_path=boot_files.initramfs_path)

        api.produce(installed_kernel_info)

        # Backwards compatibility
        # Expects that the kernel nevra has the following format: <pkg_name>-<version>-<release>.<arch>
        version = '-'.join(target_kernel_nevra.split('-')[-2:])  # (-2)-th is <version>; take <version>-<release>...
        api.produce(InstalledTargetKernelVersion(version=version))
    else:
        # This is not expected, however, we are past the point that raising an exception would do any good.
        # It is better to finish the upgrade with 80% things done rather than falling into emergency mode
        api.current_logger().warning('Failed to identify package providing the target kernel.')
        pass



from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM

# rpm : the default config file
BROWSED_CONFIG = '/etc/cups/cups-browsed.conf'


# The list of macros that should be set to get the behavior
# from previous RHEL
NEW_MACROS = [
    ('LocalQueueNamingRemoteCUPS', 'RemoteName'),
    ('CreateIPPPrinterQueues', 'All')
]


def _macro_exists(path, macro):
    """
    Check if macro is in the file.

    :param str path: string representing the full path of the config file
    :param str macro: new directive to be added
    :return boolean res: macro does/does not exist in the file
    """
    with open(path, 'r') as f:
        lines = f.readlines()

    for line in lines:
        if line.lstrip().startswith(macro):
            return True
    return False


def _append_string(path, content):
    """
    Append string at the end of file.

    :param str path: string representing the full path of file
    :param str content: preformatted string to be added
    """
    with open(path, 'a') as f:
        f.write(content)


def update_config(path, check_function=_macro_exists,
                  append_function=_append_string):
    """
    Insert expected content into the file on the path if it is not
    in the file already.

    :param str path: string representing the full path of the config file
    :param func check_function: function to be used to check if string is in the file
    :param func append_function: function to be used to append string
    """

    macros = []
    for macro in NEW_MACROS:
        if not check_function(path, macro[0]):
            macros.append(' '.join(macro))

    if not macros:
        return

    fmt_input = "\n{comment_line}\n{content}\n".format(comment_line='# content added by Leapp',
                                                       content='\n'.join(macros))

    try:
        append_function(path, fmt_input)
    except IOError:
        raise IOError('Error during writing to file: {}.'.format(path))


def _check_package(pkg):
    """
    Checks if a package is installed and signed

    :param str pkg: name of package
    """
    return has_package(DistributionSignedRPM, pkg)


def update_cups_browsed(debug_log=api.current_logger().debug,
                        error_log=api.current_logger().error,
                        is_installed=_check_package,
                        append_function=_append_string,
                        check_function=_macro_exists):
    """
    Update cups-browsed configuration file

    :param func debug_log: function for debug logging
    :param func error_log: function for error logging
    :param func is_installed: checks if the package is installed
    :param func append_function: appends string into file
    :param func check_function: checks if macro is in the file
    """

    error_list = []

    if not is_installed('cups-filters'):
        return

    debug_log('Updating cups-browsed configuration file {}.'
              .format(BROWSED_CONFIG))

    try:
        update_config(BROWSED_CONFIG,
                      check_function,
                      append_function)
    except (OSError, IOError) as error:
        error_list.append((BROWSED_CONFIG, error))
    if error_list:
        error_log('The files below have not been modified '
                  '(error message included):' +
                  ''.join(['\n    - {}: {}'.format(err[0], err[1])
                          for err in error_list]))
        return



import os

from leapp import reporting
from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.models import Report, RootDirectory
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckRootSymlinks(Actor):
    """
    Check if the symlinks /bin and /lib are relative, not absolute.

    After reboot, dracut fails if the links are absolute.
    """

    name = 'check_root_symlinks'
    consumes = (RootDirectory,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        rootdir = next(self.consume(RootDirectory), None)
        if not rootdir:
            raise StopActorExecutionError('Cannot check root symlinks',
                                          details={'Problem': 'Did not receive a message with '
                                                              'root subdirectories'})
        absolute_links = [item for item in rootdir.items if item.target and os.path.isabs(item.target)]
        absolute_links_nonutf = [item for item in rootdir.invalid_items if item.target and os.path.isabs(item.target)]
        if not absolute_links and not absolute_links_nonutf:
            return

        report_fields = [
                reporting.Title('Upgrade requires links in root directory to be relative'),
                reporting.Summary(
                    'After rebooting, parts of the upgrade process can fail if symbolic links in / '
                    'point to absolute paths.\n'
                    'Please change these links to relative ones.'
                    ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups([reporting.Groups.INHIBITOR])]

        # Generate reports about absolute links presence
        rem_commands = []
        if absolute_links:
            commands = []
            for item in absolute_links:
                command = ' '.join(['ln',
                                    '-snf',
                                    os.path.relpath(item.target, '/'),
                                    os.path.join('/', item.name)])
                commands.append(command)
            rem_commands = [['sh', '-c', ' && '.join(commands)]]
        # Generate reports about non-utf8 absolute links presence
        nonutf_count = len(absolute_links_nonutf)
        if nonutf_count > 0:
            # for non-utf encoded filenames can't provide a remediation command, so will mention this fact in a hint
            rem_hint = ("{} symbolic links point to absolute paths that have non-utf8 encoding and need to be"
                        " fixed additionally".format(nonutf_count))
            report_fields.append(reporting.Remediation(hint=rem_hint, commands=rem_commands))
        else:
            report_fields.append(reporting.Remediation(commands=rem_commands))

        reporting.create_report(report_fields)



from leapp import reporting
from leapp.actors import Actor
from leapp.models import CephInfo, StorageInfo
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class InhibitWhenLuks(Actor):
    """
    Check if any encrypted partitions is in use. If yes, inhibit the upgrade process.

    Upgrading system with encrypted partition is not supported.
    """

    name = 'check_luks_and_inhibit'
    consumes = (StorageInfo, CephInfo)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        # If encrypted Ceph volumes present, check if there are more encrypted disk in lsblk than Ceph vol
        ceph_vol = []
        try:
            ceph_info = next(self.consume(CephInfo))
            if ceph_info:
                ceph_vol = ceph_info.encrypted_volumes[:]
                for storage_info in self.consume(StorageInfo):
                    for blk in storage_info.lsblk:
                        if blk.tp == 'crypt' and blk.name not in ceph_vol:
                            create_report([
                                reporting.Title('LUKS encrypted partition detected'),
                                reporting.Summary('Upgrading system with encrypted partitions is not supported'),
                                reporting.Severity(reporting.Severity.HIGH),
                                reporting.Groups([reporting.Groups.BOOT, reporting.Groups.ENCRYPTION]),
                                reporting.Groups([reporting.Groups.INHIBITOR]),
                            ])
                            break
        except StopIteration:
            for storage_info in self.consume(StorageInfo):
                for blk in storage_info.lsblk:
                    if blk.tp == 'crypt':
                        create_report([
                            reporting.Title('LUKS encrypted partition detected'),
                            reporting.Summary('Upgrading system with encrypted partitions is not supported'),
                            reporting.Severity(reporting.Severity.HIGH),
                            reporting.Groups([reporting.Groups.BOOT, reporting.Groups.ENCRYPTION]),
                            reporting.Groups([reporting.Groups.INHIBITOR]),
                        ])
                        break



import warnings

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import module as module_lib
from leapp.libraries.common import rpms
from leapp.libraries.stdlib import api
from leapp.models import InstalledRPM, RPM

no_yum = False
no_yum_warning_msg = "package `yum` is unavailable"
try:
    import yum
except ImportError:
    no_yum = True
    warnings.warn(no_yum_warning_msg, ImportWarning)

no_dnf = False
no_dnf_warning_msg = "package `dnf` is unavailable"
try:
    import dnf
except ImportError:
    no_dnf = True
    warnings.warn(no_dnf_warning_msg, ImportWarning)


def _get_package_repository_data_yum():
    yum_base = yum.YumBase()
    pkg_repos = {}

    try:
        for pkg in yum_base.doPackageLists().installed:
            pkg_repos[pkg.name] = pkg.ui_from_repo.lstrip('@')
    except ValueError as e:
        if 'locale' not in str(e):  # reraise if error is not related to locales
            raise e
        raise StopActorExecutionError(
            message='Failed to get installed RPM packages because of an invalid locale',
            details={
                'hint': 'Please run leapp with a valid locale. ' +
                        'You can get a list of installed locales by running `locale -a`.'
            })

    return pkg_repos


def _get_package_repository_data_dnf():
    dnf_base = dnf.Base()
    pkg_repos = {}

    try:
        # NOTE: currently we do not initialize/load DNF plugins here as we are
        # working just with the local stuff (load_system_repo=True)
        dnf_base.fill_sack(load_system_repo=True, load_available_repos=False)
        for pkg in dnf_base.sack.query():
            pkg_repos[pkg.name] = pkg._from_repo.lstrip('@')
    except ValueError as e:
        if 'locale' not in str(e):  # reraise if error is not related to locales
            raise e
        raise StopActorExecutionError(
            message='Failed to get installed RPM packages because of an invalid locale',
            details={
                'hint': 'Please run leapp with a valid locale. ' +
                        'You can get a list of installed locales by running `locale -a`.'
            })

    return pkg_repos


def get_package_repository_data():
    """ Return dictionary mapping package name with repository from which it was installed.
    Note:
        There's no yum module for py3. The dnf module can be used only on RHEL 8+,
        on RHEL 7 there's a bug in dnf preventing us to do so:
        https://bugzilla.redhat.com/show_bug.cgi?id=1789840
    """
    if not no_yum:
        return _get_package_repository_data_yum()
    if not no_dnf:
        return _get_package_repository_data_dnf()
    raise StopActorExecutionError(message=no_yum_warning_msg)


def map_modular_rpms_to_modules():
    """
    Map modular packages to the module streams they come from.
    """
    modules = module_lib.get_modules()
    # empty on RHEL 7 because of no modules
    if not modules:
        return {}
    # create a reverse mapping from the RPMS to module streams
    # key: tuple of 4 strings representing a NVRA (name, version, release, arch) of an RPM
    # value: tuple of 2 strings representing a module and its stream
    rpm_streams = {}
    for module in modules:
        for rpm in module.getArtifacts():
            # we transform the NEVRA string into a tuple
            name, epoch_version, release_arch = rpm.rsplit('-', 2)
            epoch, version = epoch_version.split(':', 1)
            release, arch = release_arch.rsplit('.', 1)
            rpm_key = (name, epoch, version, release, arch)
            # stream could be int or float, convert it to str just in case
            rpm_streams[rpm_key] = (module.getName(), str(module.getStream()))
    return rpm_streams


# TODO(drehak) unit tests
def process():
    output = rpms.get_installed_rpms()
    pkg_repos = get_package_repository_data()
    rpm_streams = map_modular_rpms_to_modules()

    result = InstalledRPM()
    for entry in output:
        entry = entry.strip()
        if not entry:
            continue
        name, version, release, epoch, packager, arch, pgpsig = entry.split('|')
        repository = pkg_repos.get(name, '')
        rpm_key = (name, epoch, version, release, arch)
        module, stream = rpm_streams.get(rpm_key, (None, None))
        result.items.append(RPM(
            name=name,
            version=version,
            epoch=epoch,
            packager=packager,
            arch=arch,
            release=release,
            pgpsig=pgpsig,
            repository=repository,
            module=module,
            stream=stream))
    api.produce(result)



import logging
import subprocess

import pytest

from leapp.libraries.actor import sctpupdate
from leapp.libraries.stdlib import CalledProcessError

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    (
        'conf_content',
        'exp_new_conf_content',
        'log_should_contain',
        'log_shouldnt_contain',
        'conf_files_exists',
        'should_raise_exc',
        'logger_level',
    ),
    [
        # testing normal behaviour
        (
            'blacklist sctp',
            '#blacklist sctp',
            'Enabled SCTP',
            None,
            True,
            None,
            logging.INFO,
        ),
        # testing if regex works also in case sctp just a part of a string
        (
            'blacklist some-sctp',
            '#blacklist some-sctp',
            'Enabled SCTP',
            None,
            True,
            None,
            logging.INFO,
        ),
        # testing if script skips non sctp lines
        (
            'blacklist tcp',
            'blacklist tcp',
            'Enabled SCTP',
            None,
            True,
            None,
            logging.INFO,
        ),
        # testing if the logger message is empty on warning level
        (
            'blacklist tcp',
            'blacklist tcp',
            '',
            None,
            True,
            None,
            logging.WARNING,
        ),
        # testing if CalledProcessError raised when sed exits with non 0 and
        # logger not emits Enabled SCTP (what we want)
        (
            'blacklist tcp',
            'blacklist tcp',
            None,
            'Enabled SCTP',
            False,
            CalledProcessError,
            logging.INFO,
        ),
    ],
)
def test_enable_sctp(
    conf_content,
    exp_new_conf_content,
    log_should_contain,
    log_shouldnt_contain,
    conf_files_exists,
    should_raise_exc,
    logger_level,
    monkeypatch,
    tmpdir,
    caplog,
    capsys,
):
    def mock_run(args):
        logger.info('Calling run with %r', args)
        res = subprocess.call(args)
        if res != 0:
            raise CalledProcessError(
                message='Sed fails with error code {!r}'.format(res),
                command=args,
                result=res,
            )

    monkeypatch.setattr(sctpupdate, 'run', mock_run)

    sctp_diag_blacklist_conf = tmpdir.join('sctp_diag-blacklist.conf')
    sctp_blacklist_conf = tmpdir.join('sctp-blacklist.conf')
    if conf_files_exists:
        sctp_diag_blacklist_conf.write(conf_content)
        sctp_blacklist_conf.write(conf_content)

    with caplog.at_level(logger_level):
        if not should_raise_exc:
            sctpupdate.enable_sctp(_modprobe_d_path=str(tmpdir))
            with open(str(sctp_blacklist_conf)) as conf:
                assert conf.readlines() == [exp_new_conf_content]
        else:
            with pytest.raises(should_raise_exc):
                sctpupdate.enable_sctp(_modprobe_d_path=str(tmpdir))

    if log_should_contain is not None:
        assert log_should_contain in caplog.text
    if log_shouldnt_contain is not None:
        assert log_shouldnt_contain not in caplog.text



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import PkgManagerInfo, RHUIInfo


def handle_etc_releasever():

    target_version = api.current_actor().configuration.version.target
    reporting.create_report([
        reporting.Title(
            'Release version in /etc/dnf/vars/releasever will be set to the current target release'
        ),
        reporting.Summary(
            'On this system, Leapp detected "releasever" variable is either configured through DNF/YUM configuration '
            'file and/or the system is using RHUI infrastructure. In order to avoid issues with repofile URLs '
            '(when --release option is not provided) in cases where there is the previous major.minor version value '
            'in the configuration, release version will be set to the target release version ({}). This will also '
            'ensure the system stays on the expected target version after the upgrade'.format(target_version)
        ),
        reporting.Severity(reporting.Severity.INFO),
        reporting.Groups([reporting.Groups.UPGRADE_PROCESS]),
    ])


def process():
    pkg_facts = next(api.consume(PkgManagerInfo), None)
    rhui_facts = next(api.consume(RHUIInfo), None)
    if pkg_facts and pkg_facts.etc_releasever is not None or rhui_facts:
        handle_etc_releasever()
    else:
        api.current_logger().debug(
            'Skipping execution. "releasever" is not set in DNF/YUM vars directory and no RHUIInfo has '
            'been produced'
        )



import glob
import os

from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import DistributionSignedRPM, DynamicLinkerConfiguration, LDConfigFile, MainLDConfigFile

LD_SO_CONF_DIR = '/etc/ld.so.conf.d'
LD_SO_CONF_MAIN = '/etc/ld.so.conf'
LD_SO_CONF_DEFAULT_INCLUDE = 'ld.so.conf.d/*.conf'
LD_SO_CONF_COMMENT_PREFIX = '#'
LD_LIBRARY_PATH_VAR = 'LD_LIBRARY_PATH'
LD_PRELOAD_VAR = 'LD_PRELOAD'


def _read_file(file_path):
    with open(file_path, 'r') as fd:
        return fd.readlines()


def _is_modified(config_path):
    """ Decide if the configuration file was modified based on the package it belongs to. """
    result = run(['rpm', '-Vf', config_path], checked=False)
    if not result['exit_code']:
        return False
    modification_flags = result['stdout'].split(' ', 1)[0]
    # The file is considered modified only when the checksum does not match
    return '5' in modification_flags


def _is_included_config_custom(config_path):
    if not os.path.isfile(config_path):
        return False

    # Check if the config file has any lines that have an effect on dynamic linker configuration
    has_effective_line = False
    for line in _read_file(config_path):
        line = line.strip()
        if line and not line.startswith(LD_SO_CONF_COMMENT_PREFIX):
            has_effective_line = True
            break

    if not has_effective_line:
        return False

    is_custom = False
    try:
        package_name = run(['rpm', '-qf', '--queryformat', '%{NAME}', config_path])['stdout']
        is_custom = not has_package(DistributionSignedRPM, package_name) or _is_modified(config_path)
    except CalledProcessError:
        is_custom = True

    return is_custom


def _parse_main_config():
    """
    Extracts included configs from the main dynamic linker configuration file (/etc/ld.so.conf)
    along with lines that are likely custom. The lines considered custom are simply those that are
    not includes.

    :returns: tuple containing all the included files and lines considered custom
    :rtype:  tuple(list, list)
    """
    config = _read_file(LD_SO_CONF_MAIN)

    included_configs = []
    other_lines = []
    for line in config:
        line = line.strip()
        if line.startswith('include'):
            cfg_glob = line.split(' ', 1)[1].strip()
            cfg_glob = os.path.join('/etc', cfg_glob) if not os.path.isabs(cfg_glob) else cfg_glob
            included_configs.append(cfg_glob)
        elif line and not line.startswith(LD_SO_CONF_COMMENT_PREFIX):
            other_lines.append(line)

    return included_configs, other_lines


def scan_dynamic_linker_configuration():
    included_configs, other_lines = _parse_main_config()

    is_default_include_present = '/etc/' + LD_SO_CONF_DEFAULT_INCLUDE in included_configs
    if not is_default_include_present:
        api.current_logger().debug('The default include "{}" is not present in '
                                   'the {} file.'.format(LD_SO_CONF_DEFAULT_INCLUDE, LD_SO_CONF_MAIN))

    if is_default_include_present and len(included_configs) != 1:
        # The additional included configs will most likely be created manually by the user
        # and therefore will get flagged as custom in the next part of this function
        api.current_logger().debug('The default include "{}" is not the only include in '
                                   'the {} file.'.format(LD_SO_CONF_DEFAULT_INCLUDE, LD_SO_CONF_MAIN))

    main_config_file = MainLDConfigFile(path=LD_SO_CONF_MAIN, modified=any(other_lines), modified_lines=other_lines)

    # Expand the config paths from globs and ensure uniqueness of resulting paths
    config_paths = set()
    for cfg_glob in included_configs:
        for cfg in glob.glob(cfg_glob):
            config_paths.add(cfg)

    included_config_files = []
    for config_path in config_paths:
        config_file = LDConfigFile(path=config_path, modified=_is_included_config_custom(config_path))
        included_config_files.append(config_file)

    # Check if dynamic linker variables used for specifying custom libraries are set
    variables = [LD_LIBRARY_PATH_VAR, LD_PRELOAD_VAR]
    used_variables = [var for var in variables if os.getenv(var, None)]

    configuration = DynamicLinkerConfiguration(main_config=main_config_file,
                                               included_configs=included_config_files,
                                               used_variables=used_variables)

    if other_lines or any([config.modified for config in included_config_files]) or used_variables:
        api.produce(configuration)



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class GpgKey(Model):
    """
    GPG Public key

    It is represented by a record in the RPM DB or by a file in directory with trusted keys (or both).
    """
    topic = SystemFactsTopic
    fingerprint = fields.String()
    rpmdb = fields.Boolean()
    filename = fields.Nullable(fields.String())


class TrustedGpgKeys(Model):
    topic = SystemFactsTopic
    items = fields.List(fields.Model(GpgKey), default=[])



from __future__ import print_function

import sys

import gi

gi.require_version('NM', '1.0')
from gi.repository import NM  # noqa: E402; pylint: disable=wrong-import-position


def is_hexstring(s):
    arr = s.split(':')
    for a in arr:
        if len(a) != 1 and len(a) != 2:
            return False
        try:
            int(a, 16)
        except ValueError:
            return False
    return True


client = NM.Client.new(None)
if not client:
    print('Cannot create NM client instance')
    sys.exit(0)

processed = 0
changed = 0
errors = 0

for c in client.get_connections():
    s_ip4 = c.get_setting_ip4_config()
    processed += 1
    if s_ip4 is not None:
        client_id = s_ip4.get_dhcp_client_id()
        if client_id is not None:
            if not is_hexstring(client_id):
                new_client_id = ':'.join(hex(ord(x))[2:] for x in client_id)
                s_ip4.set_property(NM.SETTING_IP4_CONFIG_DHCP_CLIENT_ID, new_client_id)
                success = c.commit_changes(True, None)
                if success:
                    changed += 1
                else:
                    errors += 1
                print('Connection {}: \'{}\' -> \'{}\' ({})'.format(c.get_uuid(),
                                                                    client_id, new_client_id,
                                                                    'OK' if success else 'FAIL'))

print("{} processed, {} changed, {} errors".format(processed, changed, errors))



import pytest

from leapp.libraries.actor.cupsmigrate import migrate_digest


class MockFileSystem(object):
    def __init__(self,
                 infiles):
        self.files = {}
        self.ssl_dir = []
        self.files = infiles
        for path in infiles.keys():
            if path.startswith('ssl') or path.startswith('/etc/cups/ssl'):
                self.ssl_dir.append(path.rsplit('/', 1)[1])

    def readlines(self, path):
        if path in self.files.keys():
            return self.files[path].splitlines(True)
        raise IOError('Error when reading file {} - file '
                      'does not exist.'.format(path))

    def write(self, path, mode, content):
        if isinstance(content, list):
            content = ''.join(content)

        if mode == 'w':
            self.files[path] = content
        else:
            self.files[path] += content

    def copy_to_ssl(self, oldpath):
        self.ssl_dir.append(oldpath.rsplit('/', 1)[1])


testdata = (
    (
        {
            '/etc/cups/cupsd.conf': 'ifdfdfgfg\n'
        },
        'ifdfdfgfg\n'
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'AuthType Basic\nHello world\n',
        },
        'AuthType Basic\nHello world\n'
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'DefaultAuthType Negotiate\nHello world\n',
        },
        'DefaultAuthType Negotiate\nHello world\n'
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'AuthType Digest\nHello world\n',
        },
        'AuthType Basic\nHello world\n'
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'DefaultAuthType Digest\nHello world\n',
        },
        'DefaultAuthType Basic\nHello world\n'
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'DefaultAuthType BasicDigest\nHello world\n',
        },
        'DefaultAuthType Basic\nHello world\n'
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'AuthType BasicDigest\nHello world\n',
        },
        'AuthType Basic\nHello world\n'
    ),
    (
        {
            '/etc/cups/cupsd.conf': '#AuthType BasicDigest\nHello world\n',
        },
        '#AuthType BasicDigest\nHello world\n'
    )
)


@pytest.mark.parametrize('files,expected', testdata)
def test_migrate_digest(files, expected):
    op = MockFileSystem(infiles=files)

    migrate_digest(op)

    assert op.files.get('/etc/cups/cupsd.conf', None) == expected



import os
import re

from leapp.libraries.common.config import architecture, version
from leapp.models import GrubConfigError


def is_grubenv_corrupted(conf_file):
    # grubenv can be missing
    if not os.path.exists(conf_file):
        return False
    # ignore when /boot/grub2/grubenv is a symlink to its EFI counterpart
    if os.path.islink(conf_file) and os.readlink(conf_file) == '../efi/EFI/redhat/grubenv':
        return False
    with open(conf_file, 'r') as config:
        config_contents = config.read()
    return len(config_contents) != 1024 or config_contents[-1] == '\n'


def _get_config_contents(config_path):
    if os.path.isfile(config_path):
        with open(config_path, 'r') as config:
            return config.read()
    return ''


def is_grub_config_missing_final_newline(conf_file):
    config_contents = _get_config_contents(conf_file)
    return config_contents and config_contents[-1] != '\n'


def detect_config_error(conf_file):
    """
    Check grub configuration for syntax error in GRUB_CMDLINE_LINUX value.

    :return: Function returns True if error was detected, otherwise False.
    """
    with open(conf_file, 'r') as f:
        config = f.read()

    pattern = r'GRUB_CMDLINE_LINUX="[^"]+"(?!(\s*$)|(\s+(GRUB|#)))'
    return re.search(pattern, config) is not None


def scan():
    errors = []
    # Check for corrupted grubenv
    if not architecture.matches_architecture(architecture.ARCH_S390X):
        configs = ['/boot/grub2/grubenv', '/boot/efi/EFI/redhat/grubenv']
        corrupted = []
        for cfg in configs:
            if is_grubenv_corrupted(cfg):
                corrupted.append(cfg)
        if corrupted:
            errors.append(GrubConfigError(error_type=GrubConfigError.ERROR_CORRUPTED_GRUBENV, files=corrupted))

    config = '/etc/default/grub'
    # Check for GRUB_CMDLINE_LINUX syntax errors
    # XXX FIXME(ivasilev) Can we make this check a common one? For now let's limit it to rhel7->rhel8 only
    if version.get_source_major_version() == '7':
        if not architecture.matches_architecture(architecture.ARCH_S390X):
            # For now, skip just s390x, that's only one that is failing now
            # because ZIPL is used there
            if detect_config_error(config):
                errors.append(GrubConfigError(error_detected=True, files=[config],
                                              error_type=GrubConfigError.ERROR_GRUB_CMDLINE_LINUX_SYNTAX))

    # Check for missing newline errors
    if is_grub_config_missing_final_newline(config):
        errors.append(GrubConfigError(error_detected=True, error_type=GrubConfigError.ERROR_MISSING_NEWLINE,
                      files=[config]))

    return errors
